import logging
import json
import os
import shutil
from dataclasses import dataclass

from pathlib import Path
from typing import (
    Any, AnyStr, List, Dict, Union)

import jsonschema  # type: ignore

from opentrons.protocols.api_support.util import ModifiedList
from opentrons.calibration_storage import helpers, modify
from opentrons.protocols.context.labware import \
    AbstractLabware
from opentrons.types import Point
from opentrons_shared_data import load_shared_data, get_shared_data_root
from opentrons.protocols.geometry.deck_item import DeckItem
from opentrons.protocols.api_support.constants import (
    OPENTRONS_NAMESPACE, CUSTOM_NAMESPACE, STANDARD_DEFS_PATH, USER_DEFS_PATH)
from opentrons_shared_data.labware.dev_types import LabwareDefinition


MODULE_LOG = logging.getLogger(__name__)


def get_labware_definition(
    load_name: str,
    namespace: str = None,
    version: int = None,
    bundled_defs: Dict[str, LabwareDefinition] = None,
    extra_defs: Dict[str, LabwareDefinition] = None
) -> LabwareDefinition:
    """
    Look up and return a definition by load_name + namespace + version and
        return it or raise an exception

    :param str load_name: corresponds to 'loadName' key in definition
    :param str namespace: The namespace the labware definition belongs to.
        If unspecified, will search 'opentrons' then 'custom_beta'
    :param int version: The version of the labware definition. If unspecified,
        will use version 1.
    :param bundled_defs: A bundle of labware definitions to exlusively use for
        finding labware definitions, if specified
    :param extra_defs: An extra set of definitions (in addition to the system
        definitions) in which to search
    """
    load_name = load_name.lower()

    if bundled_defs is not None:
        return _get_labware_definition_from_bundle(
            bundled_defs, load_name, namespace, version)

    checked_extras = extra_defs or {}

    try:
        return _get_labware_definition_from_bundle(
            checked_extras, load_name, namespace, version)
    except (FileNotFoundError, RuntimeError):
        pass

    return _get_standard_labware_definition(
        load_name, namespace, version)


def get_all_labware_definitions() -> List[str]:
    """
    Return a list of standard and custom labware definitions with load_name +
        name_space + version existing on the robot
    """
    labware_list = ModifiedList()

    def _check_for_subdirectories(path):
        with os.scandir(path) as top_path:
            for sub_dir in top_path:
                if sub_dir.is_dir():
                    labware_list.append(sub_dir.name)

    # check for standard labware
    _check_for_subdirectories(get_shared_data_root() / STANDARD_DEFS_PATH)

    # check for custom labware
    for namespace in os.scandir(USER_DEFS_PATH):
        _check_for_subdirectories(namespace)

    return labware_list


def save_definition(
    labware_def: LabwareDefinition,
    force: bool = False,
    location: Path = None
) -> None:
    """
    Save a labware definition

    :param labware_def: A deserialized JSON labware definition
    :param bool force: If true, overwrite an existing definition if found.
        Cannot overwrite Opentrons definitions.
    :param location: File path
    """
    namespace = labware_def['namespace']
    load_name = labware_def['parameters']['loadName']
    version = labware_def['version']

    verify_definition(labware_def)

    if not namespace or not load_name or not version:
        raise RuntimeError(
            'Could not save definition, labware def is missing a field: ' +
            f'{namespace}, {load_name}, {version}')

    if namespace == OPENTRONS_NAMESPACE:
        raise RuntimeError(
            f'Saving definitions to the "{OPENTRONS_NAMESPACE}" namespace ' +
            'is not permitted')

    def_path = _get_path_to_labware(load_name, namespace, version, location)

    if not force and def_path.is_file():
        raise RuntimeError(
            f'The given definition ({namespace}/{load_name} v{version}) ' +
            'already exists. Cannot save definition without force=True')

    Path(def_path).parent.mkdir(parents=True, exist_ok=True)
    with open(def_path, 'w') as f:
        json.dump(labware_def, f)


def verify_definition(contents: Union[
        AnyStr, LabwareDefinition, Dict[str, Any]])\
        -> LabwareDefinition:
    """ Verify that an input string is a labware definition and return it.

    If the definition is invalid, an exception is raised; otherwise parse the
    json and return the valid definition.

    :raises json.JsonDecodeError: If the definition is not valid json
    :raises jsonschema.ValidationError: If the definition is not valid.
    :returns: The parsed definition
    """
    schema_body = load_shared_data('labware/schemas/2.json').decode('utf-8')
    labware_schema_v2 = json.loads(schema_body)

    if isinstance(contents, dict):
        to_return = contents
    else:
        to_return = json.loads(contents)
    jsonschema.validate(to_return, labware_schema_v2)
    # we can type ignore this because if it passes the jsonschema it has
    # the correct structure
    return to_return  # type: ignore


def delete_all_custom_labware() -> None:
    """Delete all custom labware"""
    if USER_DEFS_PATH.is_dir():
        shutil.rmtree(USER_DEFS_PATH)


def save_calibration(
        labware: AbstractLabware,
        delta: Point) -> None:
    """Save a calibration"""
    index_info = IndexFileInformation.from_labware(labware)
    modify.save_labware_calibration(
        labware_path=index_info.path,
        definition=index_info.definition,
        delta=delta,
        parent=index_info.parent
    )
    labware.set_calibration(delta=delta)


def _get_labware_definition_from_bundle(
    bundled_labware: Dict[str, LabwareDefinition],
    load_name: str,
    namespace: str = None,
    version: int = None,
) -> LabwareDefinition:
    """
    Look up and return a bundled definition by ``load_name`` + ``namespace``
    + ``version`` and return it or raise an exception. The``namespace`` and
    ``version`` args are optional, they only have to be specified if there is
    ambiguity (eg when multiple labware in the bundle share the same
    ``load_name``)

    :param str load_name: corresponds to 'loadName' key in definition
    :param str namespace: The namespace the labware definition belongs to
    :param int version: The version of the labware definition
    :param Dict bundled_labware: A dictionary of labware definitions to search
    """
    load_name = load_name.lower()

    bundled_candidates = [
        b for b in bundled_labware.values()
        if b['parameters']['loadName'] == load_name]
    if namespace:
        namespace = namespace.lower()
        bundled_candidates = [
            b for b in bundled_candidates if b['namespace'] == namespace]
    if version:
        bundled_candidates = [
            b for b in bundled_candidates if b['version'] == version]

    if len(bundled_candidates) == 1:
        return bundled_candidates[0]
    elif len(bundled_candidates) > 1:
        raise RuntimeError(
            f'Ambiguous labware access. Bundle contains multiple '
            f'labware with load name {load_name}, '
            f'namespace {namespace}, and version {version}.')
    else:
        raise RuntimeError(
            f'No labware found in bundle with load name {load_name}, '
            f'namespace {namespace}, and version {version}.')


def _get_standard_labware_definition(
        load_name: str,
        namespace: str = None,
        version: int = None) -> LabwareDefinition:

    if version is None:
        checked_version = 1
    else:
        checked_version = version
    error_msg_string = """Unable to find a labware
        definition for "{0}",
        version {1}, in the {2} namespace.
        Please confirm your protocol includes the correct
        labware spelling and (optionally) the correct version
        number and namespace.

        If you are referencing a custom labware in your
        protocol, you must add it to your Custom Labware
        Definitions Folder from the Opentrons App before
        uploading your protocol.
        """

    if namespace is None:
        for fallback_namespace in [OPENTRONS_NAMESPACE, CUSTOM_NAMESPACE]:
            try:
                return _get_standard_labware_definition(
                    load_name, fallback_namespace, checked_version)
            except FileNotFoundError:
                pass

        raise FileNotFoundError(error_msg_string.format(
                load_name, checked_version, OPENTRONS_NAMESPACE))

    namespace = namespace.lower()
    def_path = _get_path_to_labware(load_name, namespace, checked_version)

    try:
        with open(def_path, 'rb') as f:
            labware_def = json.loads(f.read().decode('utf-8'))
    except FileNotFoundError:
        raise FileNotFoundError(
            f'Labware "{load_name}" not found with version {checked_version} '
            f'in namespace "{namespace}".'
        )

    return labware_def


def _get_parent_identifier(labware: AbstractLabware) -> str:
    """
    Helper function to return whether a labware is on top of a
    module or not.
    """
    parent = labware.get_geometry().parent.labware.object
    # TODO (lc, 07-14-2020): Once we implement calibrations per slot,
    # this function should either return a slot using `first_parent` or
    # the module it is attached to.
    if isinstance(parent, DeckItem) and parent.separate_calibration:
        # treat a given labware on a given module type as same
        return parent.load_name
    else:
        return ''  # treat all slots as same


def get_labware_hash(labware: AbstractLabware) -> str:
    return helpers.hash_labware_def(labware.get_definition())


def get_labware_hash_with_parent(labware: AbstractLabware) -> str:
    return helpers.hash_labware_def(
        labware.get_definition()
    ) + _get_parent_identifier(labware)


def _get_labware_path(labware: AbstractLabware) -> str:
    return f'{get_labware_hash_with_parent(labware)}.json'


@dataclass(frozen=True)
class IndexFileInformation:
    definition: LabwareDefinition
    parent: str
    path: str

    @classmethod
    def from_labware(cls, labware: AbstractLabware) -> 'IndexFileInformation':
        return IndexFileInformation(
            definition=labware.get_definition(),
            path=_get_labware_path(labware),
            parent=_get_parent_identifier(labware)
        )


def _get_path_to_labware(
        load_name: str, namespace: str, version: int, base_path: Path = None
        ) -> Path:
    if namespace == OPENTRONS_NAMESPACE:
        # all labware in OPENTRONS_NAMESPACE is stored in shared data
        return get_shared_data_root() / STANDARD_DEFS_PATH \
               / load_name / f'{version}.json'
    if not base_path:
        base_path = USER_DEFS_PATH
    def_path = base_path / namespace / load_name / f'{version}.json'
    return def_path



from __future__ import annotations

from typing import Optional, cast, TYPE_CHECKING


from opentrons.types import Point
from opentrons_shared_data.labware.dev_types import (
    WellDefinition, CircularWellDefinition, RectangularWellDefinition)

if TYPE_CHECKING:
    from opentrons.protocols.context.labware import \
        AbstractLabware


class WellGeometry:

    def __init__(self,
                 well_props: WellDefinition,
                 parent_point: Point,
                 parent_object: AbstractLabware):
        """
        Construct a well geometry object.

        :param well_props: Properties from the labware definition
        :param parent_point: The coordinate of parent labware
        :param parent_object: The parent labware
        """

        self._position\
            = Point(well_props['x'],
                    well_props['y'],
                    well_props['z'] + well_props['depth']) + parent_point

        if not parent_object:
            raise ValueError("Wells must have a parent")

        self._parent = parent_object

        self._length: Optional[float] = None
        self._width: Optional[float] = None
        self._diameter: Optional[float] = None

        shape = well_props['shape']
        if shape == 'rectangular':
            rect_props = cast(RectangularWellDefinition, well_props)
            self._length = rect_props['xDimension']
            self._width = rect_props['yDimension']
            self._x_size = self._length
            self._y_size = self._width
        elif shape == 'circular':
            circular_props = cast(CircularWellDefinition, well_props)
            self._diameter = circular_props['diameter']
            self._x_size = self._y_size = self._diameter
        else:
            raise ValueError(
                f'Shape "{shape}" is not a supported well shape')

        self._max_volume = well_props['totalLiquidVolume']
        self._depth = well_props['depth']

    @property
    def parent(self) -> AbstractLabware:
        return self._parent

    @property
    def position(self) -> Point:
        return self._position

    @property
    def diameter(self) -> Optional[float]:
        return self._diameter

    def top(self, z: float = 0.0) -> Point:
        return self._position + Point(0, 0, z)

    def bottom(self, z: float = 0.0) -> Point:
        top = self.top()
        bottom_z = top.z - self._depth + z
        return Point(x=top.x, y=top.y, z=bottom_z)

    def center(self) -> Point:
        top = self.top()
        center_z = top.z - (self._depth / 2.0)
        return Point(x=top.x, y=top.y, z=center_z)

    @property
    def max_volume(self) -> float:
        return self._max_volume

    def from_center_cartesian(
            self, x: float, y: float, z: float) -> Point:
        """
        Specifies an arbitrary point in deck coordinates based
        on percentages of the radius in each axis. For example, to specify the
        back-right corner of a well at 1/4 of the well depth from the bottom,
        the call would be `from_center_cartesian(1, 1, -0.5)`.

        No checks are performed to ensure that the resulting position will be
        inside of the well.

        To use this point in a liquid handling command, you must create a
        location object using the well such as,

        .. code-block:: python

            from opentrons import types

            # Cannot be passed into a liquid handling function
            point = well.from_center_cartesian(1, 1, -0.5)

            # Can be passed into a liquid handling function
            location = types.Location(
                well.from_center_cartesian(1, 1, -0.5), well)

        :param x: a float in the range [-1.0, 1.0] for a percentage of half of
            the radius/length in the X axis
        :param y: a float in the range [-1.0, 1.0] for a percentage of half of
            the radius/width in the Y axis
        :param z: a float in the range [-1.0, 1.0] for a percentage of half of
            the height above/below the center

        :return: a Point representing the specified location in absolute deck
        coordinates
        """
        center = self.center()
        x_size = self._x_size
        y_size = self._y_size
        z_size = self._depth

        return Point(
            x=center.x + (x * (x_size / 2.0)),
            y=center.y + (y * (y_size / 2.0)),
            z=center.z + (z * (z_size / 2.0)))



"""
opentrons.protocols.parse: functions and state for parsing protocols
"""

import ast
import functools
import itertools
import json
import logging
import re
from io import BytesIO
from zipfile import ZipFile
from typing import Any, Dict, Optional, Union, Tuple, TYPE_CHECKING

import jsonschema  # type: ignore

from opentrons.config import feature_flags as ff
from opentrons_shared_data import load_shared_data, protocol
from .api_support.types import APIVersion
from .types import (Protocol, PythonProtocol, JsonProtocol,
                    Metadata, MalformedProtocolError,
                    ApiDeprecationError)
from .bundle import extract_bundle

if TYPE_CHECKING:
    from opentrons_shared_data.labware.dev_types import LabwareDefinition
    from opentrons_shared_data.protocol.dev_types import (
        JsonProtocol as JsonProtocolDef
    )

MODULE_LOG = logging.getLogger(__name__)

# match e.g. "2.0" but not "hi", "2", "2.0.1"
API_VERSION_RE = re.compile(r'^(\d+)\.(\d+)$')
MAX_SUPPORTED_JSON_SCHEMA_VERSION = 5
API_VERSION_FOR_JSON_V5_AND_BELOW = APIVersion(2, 8)


def _validate_v2_ast(protocol_ast: ast.Module):
    defs = [fdef for fdef in protocol_ast.body
            if isinstance(fdef, ast.FunctionDef)]
    rundefs = [fdef for fdef in defs
               if fdef.name == 'run']
    # There must be precisely 1 one run function
    if len(rundefs) > 1:
        lines = [str(d.lineno) for d in rundefs]
        linestr = ', '.join(lines)
        raise MalformedProtocolError(
            f'More than one run function is defined (lines {linestr})')
    if not rundefs:
        raise MalformedProtocolError(
            "No function 'run(ctx)' defined")
    if infer_version_from_imports(protocol_ast):
        raise MalformedProtocolError(
            'Protocol API v1 modules such as robot, instruments, and labware '
            'may not be imported in Protocol API V2 protocols'
        )


def version_from_string(vstr: str) -> APIVersion:
    """ Parse an API version from a string

    :param str vstr: The version string to parse
    :returns APIVersion: The parsed version
    :raises ValueError: if the version string is the wrong format
    """
    matches = API_VERSION_RE.match(vstr)
    if not matches:
        raise ValueError(
            f'apiLevel {vstr} is incorrectly formatted. It should '
            'major.minor, where both major and minor are numbers.')
    return APIVersion(
        major=int(matches.group(1)), minor=int(matches.group(2)))


def _parse_json(
        protocol_contents: str, filename: str = None) -> JsonProtocol:
    """ Parse a protocol known or at least suspected to be json """
    protocol_json = json.loads(protocol_contents)
    version, validated = validate_json(protocol_json)
    return JsonProtocol(
        text=protocol_contents, filename=filename, contents=validated,
        schema_version=version, api_level=API_VERSION_FOR_JSON_V5_AND_BELOW,
        metadata=validated['metadata']
    )


def _parse_python(
    protocol_contents: str,
    filename: str = None,
    bundled_labware: Dict[str, 'LabwareDefinition'] = None,
    bundled_data: Dict[str, bytes] = None,
    bundled_python: Dict[str, str] = None,
    extra_labware: Dict[str, 'LabwareDefinition'] = None,
) -> PythonProtocol:
    """ Parse a protocol known or at least suspected to be python """
    filename_checked = filename or '<protocol>'
    if filename_checked.endswith('.zip'):
        ast_filename = 'protocol.ot2.py'
    else:
        ast_filename = filename_checked

    parsed = ast.parse(protocol_contents,
                       filename=ast_filename)

    metadata = extract_metadata(parsed)
    protocol = compile(parsed, filename=ast_filename, mode='exec')
    version = get_version(metadata, parsed)

    if version >= APIVersion(2, 0):
        _validate_v2_ast(parsed)
    else:
        raise ApiDeprecationError(version)

    result = PythonProtocol(
        text=protocol_contents,
        filename=getattr(protocol, 'co_filename', '<protocol>'),
        contents=protocol,
        metadata=metadata,
        api_level=version,
        bundled_labware=bundled_labware,
        bundled_data=bundled_data,
        bundled_python=bundled_python,
        extra_labware=extra_labware)

    return result


def _parse_bundle(bundle: ZipFile, filename: str = None) -> PythonProtocol:
    """ Parse a bundled Python protocol """
    if not ff.use_protocol_api_v2():
        raise RuntimeError(
            'Uploading a bundled protocol requires the robot to be set to '
            'Protocol API V2. Enable the \'Use Protocol API version 2\' '
            'toggle in the robot\'s Advanced Settings and restart the robot')

    contents = extract_bundle(bundle)

    result = _parse_python(
        contents.protocol, filename,
        contents.bundled_labware,
        contents.bundled_data,
        contents.bundled_python)

    if result.api_level < APIVersion(2, 0):
        raise RuntimeError('Bundled protocols must use Protocol API V2, ' +
                           f'got {result.api_level}')

    return result


def parse(
    protocol_file: Union[str, bytes],
    filename: str = None,
    extra_labware: Dict[str, 'LabwareDefinition'] = None,
    extra_data: Dict[str, bytes] = None
) -> Protocol:
    """ Parse a protocol from text.

    :param protocol_file: The protocol file, or for single-file protocols, a
                        string of the protocol contents.
    :param filename: The name of the protocol. Optional, but helps with
                     deducing the kind of protocol (e.g. if it ends with
                     '.json' we can treat it like json)
    :param extra_labware: Any extra labware defs that should be given to the
                          protocol. Ignored if the protocol is json or zipped
                          python.
    :param extra_data: Any extra data files that should be provided to the
                       protocol. Ignored if the protocol is json or zipped
                       python.
    :return types.Protocol: The protocol holder, a named tuple that stores the
                        data in the protocol for later simulation or
                        execution.
    """
    if filename and filename.endswith('.zip'):
        if not isinstance(protocol_file, bytes):
            raise RuntimeError('Please update your Run App version to '
                               'support uploading a .zip file')

        with ZipFile(BytesIO(protocol_file)) as bundle:
            result = _parse_bundle(bundle, filename)
        return result
    else:
        if isinstance(protocol_file, bytes):
            protocol_str = protocol_file.decode('utf-8')
        else:
            protocol_str = protocol_file

        if filename and filename.endswith('.json'):
            return _parse_json(protocol_str, filename)
        elif filename and filename.endswith('.py'):
            return _parse_python(
                protocol_str, filename, extra_labware=extra_labware,
                bundled_data=extra_data)

        # our jsonschema says the top level json kind is object
        if protocol_str and protocol_str[0] in ('{', b'{'):
            return _parse_json(protocol_str, filename)
        else:
            return _parse_python(
                protocol_str, filename, extra_labware=extra_labware,
                bundled_data=extra_data)


def extract_metadata(parsed: ast.Module) -> Metadata:
    metadata: Metadata = {}
    assigns = [
        obj for obj in parsed.body if isinstance(obj, ast.Assign)]
    for obj in assigns:
        # XXX This seems brittle and could probably do with
        # - enough love that we can remove the type: ignores
        # - some thought about exactly what types are allowed in metadata
        if isinstance(obj.targets[0], ast.Name) \
                and obj.targets[0].id == 'metadata' \
                and isinstance(obj.value, ast.Dict):
            keys = [k.s for k in obj.value.keys]  # type: ignore
            values = [v.s for v in obj.value.values]  # type: ignore
            metadata = dict(zip(keys, values))
    return metadata


@functools.lru_cache(1)
def infer_version_from_imports(parsed: ast.Module) -> Optional[APIVersion]:
    # Imports in the form of `import opentrons.robot` will have an entry in
    # parsed.body[i].names[j].name in the form "opentrons.robot". Find those
    # imports and transform them to strip away the 'opentrons.' part.
    ot_imports = ['.'.join(name.name.split('.')[1:]) for name in
                  itertools.chain.from_iterable(
                      [obj.names for obj in parsed.body
                       if isinstance(obj, ast.Import)])
                  if 'opentrons' in name.name]

    # Imports in the form of `from opentrons import robot` (with or without an
    # `as ___` statement) will have an entry in parsed.body[i].module
    # containing "opentrons"
    ot_from_imports = [
        name.name for name in
        itertools.chain.from_iterable(
            [obj.names for obj in parsed.body
             if isinstance(obj, ast.ImportFrom)
             and obj.module
             and 'opentrons' in obj.module])
    ]

    # If any of these are populated, filter for entries with v1-specific terms
    opentrons_imports = set(ot_imports + ot_from_imports)
    v1_markers = set(('robot', 'instruments', 'modules', 'containers'))
    v1evidence = v1_markers.intersection(opentrons_imports)
    if v1evidence:
        return APIVersion(1, 0)
    else:
        return None


def version_from_metadata(metadata: Metadata) -> APIVersion:
    """ Build an API version from metadata, if we can.

    If there is no apiLevel key, raise a KeyError.
    If the apiLevel value is malformed, raise a ValueError.
    """
    if 'apiLevel' not in metadata:
        raise KeyError('apiLevel')
    requested_level = str(metadata['apiLevel'])
    if requested_level == '1':
        return APIVersion(1, 0)

    return version_from_string(requested_level)


def get_version(metadata: Metadata, parsed: ast.Module) -> APIVersion:
    """
    Infer protocol API version based on a combination of metadata and imports.

    If a protocol specifies its API version using the 'apiLevel' key of a top-
    level dict variable named `metadata`, the value for that key will be
    returned as the version (the value will be intified, so numeric values
    only can be used).

    If that variable does not exist or if it does not contain the 'apiLevel'
    key, the API version will be inferred from the imports. A script with an
    import containing 'robot', 'instruments', or 'modules' will be assumed to
    be an APIv1 protocol. If none of these are present, it is assumed to be an
    APIv2 protocol (note that 'labware' is not in this list, as there is a
    valid APIv2 import named 'labware').
    """
    try:
        return version_from_metadata(metadata)
    except KeyError:  # No apiLevel key, may be apiv1
        pass
    inferred = infer_version_from_imports(parsed)
    if not inferred:
        raise RuntimeError(
            'If this is not an API v1 protocol, you must specify the target '
            'api level in the apiLevel key of the metadata. For instance, '
            'metadata={"apiLevel": "2.0"}')
    return inferred


def _get_protocol_schema_version(protocol_json: Dict[Any, Any]) -> int:
    # v3 and above uses `schemaVersion: integer`
    version = protocol_json.get('schemaVersion')
    if version:
        return int(version)
    # v1 uses 1.x.x and v2 uses 2.x.x
    legacyKebabVersion = protocol_json.get('protocol-schema')
    # No minor/patch schemas ever were released,
    # do not permit protocols with nonexistent schema versions to load
    if legacyKebabVersion == '1.0.0':
        return 1
    elif legacyKebabVersion == '2.0.0':
        return 2
    elif legacyKebabVersion:
        raise RuntimeError(
            f'No such schema version: "{legacyKebabVersion}". Did you mean ' +
            '"1.0.0" or "2.0.0"?')
    # no truthy value for schemaVersion or protocol-schema
    raise RuntimeError(
        'Could not determine schema version for protocol. ' +
        'Make sure there is a version number under "schemaVersion"')


def _get_schema_for_protocol(version_num: int) -> protocol.Schema:
    """ Retrieve the json schema for a protocol schema version
    """
    # TODO(IL, 2020/03/05): use $otSharedSchema, but maybe wait until
    # deprecating v1/v2 JSON protocols?
    if version_num > MAX_SUPPORTED_JSON_SCHEMA_VERSION:
        raise RuntimeError(
            f'JSON Protocol version {version_num} is not yet ' +
            'supported in this version of the API')
    try:
        schema = load_shared_data(
            f'protocol/schemas/{version_num}.json')
    except FileNotFoundError:
        schema = None  # type: ignore
    if not schema:
        raise RuntimeError('JSON Protocol schema "{}" does not exist'
                           .format(version_num))
    return json.loads(schema.decode('utf-8'))


def validate_json(
        protocol_json: Dict[Any, Any]) -> Tuple[int, 'JsonProtocolDef']:
    """ Validates a json protocol and returns its schema version """
    # Check if this is actually a labware
    labware_schema_v2 = json.loads(load_shared_data(
        'labware/schemas/2.json').decode('utf-8'))
    try:
        jsonschema.validate(protocol_json, labware_schema_v2)
    except jsonschema.ValidationError:
        pass
    else:
        MODULE_LOG.error("labware uploaded instead of protocol")
        raise RuntimeError(
            'The file you are trying to open is a JSON labware definition, '
            'and therefore can not be opened here. Please try '
            'uploading a JSON protocol file instead.')

    # this is now either a protocol or something corrupt
    version_num = _get_protocol_schema_version(protocol_json)
    if version_num <= 2:
        raise RuntimeError(
            f'JSON protocol version {version_num} is '
            'deprecated. Please upload your protocol into Protocol '
            'Designer and save it to migrate the protocol to a later '
            'version. This error might mean a labware '
            'definition was specified instead of a protocol.')
    if version_num > MAX_SUPPORTED_JSON_SCHEMA_VERSION:
        raise RuntimeError(
            f'The protocol you are trying to open is a JSONv{version_num} '
            'protocol and is not supported by your current robot server '
            'version. Please update your OT-2 App and robot server to the '
            'latest version and try again.'
        )
    protocol_schema = _get_schema_for_protocol(version_num)

    # instruct schema how to resolve all $ref's used in protocol schemas
    resolver = jsonschema.RefResolver(
        protocol_schema.get('$id', ''),
        protocol_schema,
        store={
            "opentronsLabwareSchemaV2": labware_schema_v2
        })

    # do the validation
    try:
        jsonschema.validate(protocol_json, protocol_schema, resolver=resolver)
    except jsonschema.ValidationError:
        MODULE_LOG.exception("JSON protocol validation failed")
        raise RuntimeError(
            'This may be a corrupted file or a JSON file that is not an '
            'Opentrons JSON protocol.')
    else:
        return version_num, protocol_json  # type: ignore



'''
Runs the XYZA axes for 24-hours, and every 10 cycles will test to see if
any of the axes have skipped by >0.5 mm

Run through an SSH session by calling with "nohup" to keep it running after
disconnecting the terminal

Example of calling this script:

    nohup python -m opentrons.tools.overnight_test &

'''

import time
import logging
from typing import Dict, List, TYPE_CHECKING

from opentrons.drivers.smoothie_drivers.driver_3_0 import \
    SmoothieError, DEFAULT_AXES_SPEED
from . import args_handler
if TYPE_CHECKING:
    from opentrons.drivers.smoothie_drivers.driver_3_0 import \
        SmoothieDriver_3_0_0


test_time_minutes = 24 * 60
start_time_minutes = int(time.time() / 60)

attempts_to_home = 1
too_many_attempts_to_home = 3

XY_TOLERANCE = 30.0
ZA_TOLERANCE = 10.0

AXIS_TEST_SKIPPING_TOLERANCE = 0.5


def setup_logging():
    # create logger with 'spam_application'
    logger = logging.getLogger('qc-test')
    logger.setLevel(logging.DEBUG)
    # create file handler which logs even debug messages
    fh = logging.FileHandler('qc-test.log')
    fh.setLevel(logging.DEBUG)
    # create console handler with a higher log level
    ch = logging.StreamHandler()
    ch.setLevel(logging.ERROR)
    # create formatter and add it to the handlers
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(fh)
    logger.addHandler(ch)
    logging.getLogger().handlers = [fh, ch]
    return logger


def attempt_movement(driver, logger, coords_list):
    driver.set_speed(DEFAULT_AXES_SPEED)
    for c in coords_list:
        try:
            driver.move(c)
        except SmoothieError:
            driver.turn_on_red_button_light()
            logger.exception('Failed movement: {}'.format(c))
            driver._reset_from_error()
            try:
                driver.update_position()
            except Exception:
                pass
            logger.info('Current Smoothie pos: {}'.format(driver.position))
            attempt_homing(driver, logger)


def attempt_homing(
        driver: 'SmoothieDriver_3_0_0', logger: logging.Logger):
    global attempts_to_home
    logger.info('Resetting Smoothieware...')
    driver._smoothie_reset()
    driver._setup()
    try:
        driver.home('XYZA')
    except SmoothieError:
        driver.turn_on_red_button_light()
        attempts_to_home += 1
        logger.exception('Failed homing')
        if attempts_to_home < too_many_attempts_to_home:
            logger.info('Attempting to home again...')
            attempt_homing(driver, logger)
            return
        else:
            raise RuntimeError(
                'Failed to home {} times, quitting the test'.format(
                    too_many_attempts_to_home))
    # if it succeeds, reset the counter to zero
    attempts_to_home = 0


def run_time_trial(
        driver: 'SmoothieDriver_3_0_0', logger: logging.Logger,
        COORDS_HOURGLASS: List[Dict[str, float]],
        COORDS_BOWTIE: List[Dict[str, float]],
        COORDS_Z_STAGE: List[Dict[str, float]]):
    while int(time.time() / 60) - start_time_minutes < test_time_minutes:
        for i in range(10):
            attempt_movement(
                driver, logger, (COORDS_HOURGLASS + COORDS_BOWTIE))
            attempt_movement(driver, logger, COORDS_Z_STAGE)
        test_all_axes(driver, logger, COORDS_MAX)
        attempt_homing(driver, logger)
        # log a message, so we get confirmation that the test is still running
        logger.info('Test is still running :)')


def test_all_axes(
        driver: 'SmoothieDriver_3_0_0', logger: logging.Logger,
        COORDS_MAX: Dict[str, float]):
    driver.move(COORDS_MAX)
    for ax in ['Z', 'A', 'X', 'Y']:
        test_axis(driver, logger, ax)


def test_axis(
        driver: 'SmoothieDriver_3_0_0', logger: logging.Logger, axis: str):
    retract_amounts = {
        'X': 3,
        'Y': 3,
        'Z': 2,
        'A': 2,
        'B': 2,
        'C': 2
    }
    retract = retract_amounts[axis]
    expected_point = driver.homed_position[axis] + retract
    points = [
        expected_point - AXIS_TEST_SKIPPING_TOLERANCE,
        expected_point + AXIS_TEST_SKIPPING_TOLERANCE
    ]
    if axis == 'Y':
        driver.move({'X': driver.homed_position['X']})
    driver.push_speed()
    driver.set_speed(8)
    driver.update_position()
    try:
        driver.move({axis: points[0]})
    except SmoothieError:
        driver.turn_on_red_button_light()
        logger.error('Test Failed: Pressing {} too soon'.format(axis))
    if axis == 'Y':
        if driver.switch_state[axis] is not False:
            driver.turn_on_red_button_light()
            logger.error('Test Failed: Pressing {} too soon'.format(axis))
        else:
            driver.move({axis: points[1]})
            if driver.switch_state[axis] is not True:
                driver.turn_on_red_button_light()
                logger.error(
                    'Test Failed: Not hitting {} switch'.format(axis))
    else:
        try:
            driver.move({axis: points[1]})
            driver.turn_on_red_button_light()
            logger.error('Test Failed: Not hitting {} switch'.format(axis))
        except SmoothieError:
            pass  # hit the switch on purpose, so it's ok
    driver.set_speed(DEFAULT_AXES_SPEED)


if __name__ == '__main__':
    logger = setup_logging()
    logger.info('Starting 24-hours Test')
    parser = args_handler.root_argparser(
        "run a long-running (overnight) series of tests for drift")
    args = parser.parse_args()
    _, driver = args_handler.build_driver(args.port)
    COORDS_MAX = {
        'X': driver.homed_position['X'] - XY_TOLERANCE,
        'Y': driver.homed_position['Y'] - XY_TOLERANCE,
        'Z': driver.homed_position['Z'] - ZA_TOLERANCE,
        'A': driver.homed_position['A'] - ZA_TOLERANCE,
    }

    COORDS_MIN = {
        'X': XY_TOLERANCE,
        'Y': XY_TOLERANCE,
        'Z': ZA_TOLERANCE,
        'A': ZA_TOLERANCE,
    }

    COORDS_MIDDLE = {
        'X': ((COORDS_MAX['X'] - COORDS_MIN['X']) / 2) + COORDS_MIN['X'],
        'Y': ((COORDS_MAX['Y'] - COORDS_MIN['Y']) / 2) + COORDS_MIN['Y'],
    }

    COORDS_HOURGLASS = [
        {'X': COORDS_MIN['X'], 'Y': COORDS_MIN['Y']},
        {'X': COORDS_MAX['X'], 'Y': COORDS_MIN['Y']},
        {'X': COORDS_MIN['X'], 'Y': COORDS_MAX['Y']},
        {'X': COORDS_MAX['X'], 'Y': COORDS_MAX['Y']}
    ]

    COORDS_BOWTIE = [
        {'X': COORDS_MAX['X'], 'Y': COORDS_MIN['Y']},
        {'X': COORDS_MIN['X'], 'Y': COORDS_MAX['Y']},
        {'X': COORDS_MIN['X'], 'Y': COORDS_MIN['Y']},
        {'X': COORDS_MAX['X'], 'Y': COORDS_MAX['Y']}
    ]

    COORDS_Z_STAGE = [
        {'X': COORDS_MIDDLE['X'], 'Y': COORDS_MIDDLE['Y']},
        {'Z': COORDS_MIN['Z'], 'A': COORDS_MIN['A']},
        {'Z': COORDS_MAX['Z'], 'A': COORDS_MAX['A']},
        {'Z': COORDS_MIN['Z'], 'A': COORDS_MIN['A']},
        {'Z': COORDS_MAX['Z'], 'A': COORDS_MAX['A']},
        {'Z': COORDS_MIN['Z'], 'A': COORDS_MIN['A']},
        {'Z': COORDS_MAX['Z'], 'A': COORDS_MAX['A']}
    ]
    try:
        driver.turn_on_green_button_light()
        attempt_homing(driver, logger)
        test_all_axes(driver, logger, COORDS_MAX)
        run_time_trial(driver, logger,
                       COORDS_HOURGLASS,
                       COORDS_BOWTIE,
                       COORDS_Z_STAGE)
    except Exception:
        driver.turn_on_red_button_light()
        logger.exception('Unexpected Error')
        exit()
    finally:
        logger.info('Exiting test')
        driver._smoothie_reset()



import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .dev_types import GPIODriverLike

MODULE_LOG = logging.getLogger(__name__)


class RevisionPinsError(Exception):
    pass


def build_gpio_chardev(chip_name: str) -> 'GPIODriverLike':
    try:
        from .gpio import GPIOCharDev
        return GPIOCharDev(chip_name)
    except (ImportError, OSError):
        MODULE_LOG.warning(
            'Failed to initialize character device, will not '
            'be able to control gpios (lights, button, smoothie'
            'kill, smoothie reset). Only one connection can be'
            ' made to the gpios at a time. '
            'If you need to control gpios, first stop the robot '
            'server with systemctl stop opentrons-robot-server. '
            'Until you restart the server with systemctl start '
            'opentrons-robot-server, you will be unable to '
            'control the robot using the Opentrons app.')
        from .gpio_simulator import SimulatingGPIOCharDev
        return SimulatingGPIOCharDev(chip_name)



import asyncio
import logging
from typing import Mapping, Optional, Union
from opentrons.drivers.mag_deck import (
    SimulatingDriver, MagDeck as MagDeckDriver)
from opentrons.drivers.mag_deck.driver import mag_locks
from opentrons.drivers.rpi_drivers.types import USBPort
from ..execution_manager import ExecutionManager
from . import update, mod_abc, types

log = logging.getLogger('__name__')

MAX_ENGAGE_HEIGHT = {  # mm from home position
    'magneticModuleV1': 45,
    'magneticModuleV2': 25}
OFFSET_TO_LABWARE_BOTTOM = {
    'magneticModuleV1': 5,
    'magneticModuleV2': 2.5}

FIRST_GEN2_REVISION = 20


def _model_from_revision(revision: Optional[str]) -> str:
    """ Defines the revision -> model mapping """
    if not revision or 'v' not in revision:
        log.error(f'bad revision: {revision}')
        return 'magneticModuleV1'
    try:
        revision_num = float(revision.split('v')[-1])  # type: ignore
    except (ValueError, TypeError):
        log.exception('bad revision: {revision}')
        return 'magneticModuleV1'
    if revision_num < FIRST_GEN2_REVISION:
        return 'magneticModuleV1'
    else:
        return 'magneticModuleV2'


class MissingDevicePortError(Exception):
    pass


class MagDeck(mod_abc.AbstractModule):
    """
    Under development. API subject to change
    """
    @classmethod
    async def build(cls,
                    port: str,
                    usb_port: USBPort,
                    execution_manager: ExecutionManager,
                    interrupt_callback: types.InterruptCallback = None,
                    simulating=False,
                    loop: asyncio.AbstractEventLoop = None,
                    sim_model: str = None):
        # MagDeck does not currently use interrupts, so the callback is not
        # passed on
        mod = cls(port=port,
                  usb_port=usb_port,
                  simulating=simulating,
                  loop=loop,
                  execution_manager=execution_manager,
                  sim_model=sim_model)
        await mod._connect()
        return mod

    @classmethod
    def name(cls) -> str:
        return 'magdeck'

    def model(self) -> str:
        return _model_from_revision(self._device_info.get('model'))

    @classmethod
    def bootloader(cls) -> types.UploadFunction:
        return update.upload_via_avrdude

    @staticmethod
    def _build_driver(
            simulating: bool,
            sim_model: str = None
    ) -> Union['SimulatingDriver', 'MagDeckDriver']:
        if simulating:
            return SimulatingDriver(sim_model=sim_model)
        else:
            return MagDeckDriver()

    def __init__(self,
                 port: str,
                 usb_port: USBPort,
                 execution_manager: ExecutionManager,
                 simulating: bool,
                 loop: asyncio.AbstractEventLoop = None,
                 sim_model: str = None) -> None:
        super().__init__(port=port,
                         usb_port=usb_port,
                         simulating=simulating,
                         loop=loop,
                         execution_manager=execution_manager,
                         sim_model=sim_model)
        self._device_info: Mapping[str, str] = {}
        self._driver: Union['SimulatingDriver', 'MagDeckDriver']
        if mag_locks.get(port):
            self._driver = mag_locks[port][1]
        else:
            self._driver = self._build_driver(
                simulating, sim_model)

    async def calibrate(self):
        """
        Calibration involves probing for top plate to get the plate height
        """
        await self.wait_for_is_running()
        self._driver.probe_plate()
        # return if successful or not?

    async def engage(self, height: float):
        """
        Move the magnet to a specific height, in mm from home position
        """
        await self.wait_for_is_running()
        if height > MAX_ENGAGE_HEIGHT[self.model()] or height < 0:
            raise ValueError(
                f'Invalid engage height for {self.model()}: {height} mm. '
                f'Must be 0 - {MAX_ENGAGE_HEIGHT[self.model()]} mm')
        self._driver.move(height)

    async def deactivate(self):
        """
        Home the magnet
        """
        await self.wait_for_is_running()
        self._driver.home()
        await self.engage(0.0)

    @property
    def current_height(self) -> float:
        return self._driver.mag_position

    @property
    def device_info(self) -> Mapping[str, str]:
        """
        Returns a dict:
            { 'serial': 'abc123', 'model': '8675309', 'version': '9001' }
        """
        return self._device_info

    @property
    def status(self) -> str:
        if self.current_height > 0:
            return 'engaged'
        else:
            return 'disengaged'

    @property
    def engaged(self) -> bool:
        if self.current_height > 0:
            return True
        else:
            return False

    @property
    def live_data(self) -> types.LiveData:
        return {
            'status': self.status,
            'data': {
                'engaged': self.engaged,
                'height': self.current_height
            }
        }

    @property
    def port(self) -> str:
        return self._port

    @property
    def usb_port(self) -> USBPort:
        return self._usb_port

    @property
    def is_simulated(self) -> bool:
        return isinstance(self._driver, SimulatingDriver)

    @property
    def interrupt_callback(self) -> types.InterruptCallback:
        return lambda x: None

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    def set_loop(self, loop: asyncio.AbstractEventLoop):
        self._loop = loop

    # Internal Methods

    async def _connect(self):
        """
        Connect to the serial port
        """
        if not self._driver.is_connected():
            self._driver.connect(self._port)
        self._device_info = self._driver.get_device_info()

    def _disconnect(self):
        """
        Disconnect from the serial port
        """
        if self._driver:
            self._driver.disconnect(port=self._port)

    def cleanup(self) -> None:
        self._disconnect()

    def __del__(self):
        self.cleanup()

    async def prep_for_update(self) -> str:
        self._driver.enter_programming_mode()
        new_port = await update.find_bootloader_port()
        return new_port or self.port



"""Command side-effect execution logic container."""
from __future__ import annotations

from opentrons.hardware_control.api import API as HardwareAPI

from ..resources import ResourceProviders
from ..state import StateView
from .equipment import EquipmentHandler
from .movement import MovementHandler
from .pipetting import PipettingHandler


class CommandHandlers:
    """CommandHandlers container class.

    CommandHandlers wraps various child handlers that define procedures to
    execute the side-effects of commands.
    """

    _equipment: EquipmentHandler
    _movement: MovementHandler
    _pipetting: PipettingHandler

    @classmethod
    def create(
        cls,
        hardware: HardwareAPI,
        state: StateView,
        resources: ResourceProviders,
    ) -> CommandHandlers:
        """Create a CommandHandlers container and its child handlers."""
        equipment = EquipmentHandler(
            state=state,
            hardware=hardware,
            resources=resources,
        )

        movement = MovementHandler(
            state=state,
            hardware=hardware
        )

        pipetting = PipettingHandler(
            state=state,
            hardware=hardware,
            movement_handler=movement,
        )

        return cls(
            equipment=equipment,
            movement=movement,
            pipetting=pipetting,
        )

    def __init__(
        self,
        equipment: EquipmentHandler,
        movement: MovementHandler,
        pipetting: PipettingHandler,
    ) -> None:
        """Initialize a CommandHandlers container."""
        self._equipment = equipment
        self._movement = movement
        self._pipetting = pipetting

    @property
    def equipment(self) -> EquipmentHandler:
        """Access equipment handling procedures."""
        return self._equipment

    @property
    def movement(self) -> MovementHandler:
        """Access movement handling procedures."""
        return self._movement

    @property
    def pipetting(self) -> PipettingHandler:
        """Access pipetting handling procedures."""
        return self._pipetting



import json
import os
import sys
from functools import partial
from pathlib import Path
import logging
import asyncio
import re
from typing import List, Tuple

from opentrons.drivers.serial_communication import get_ports_by_name
from opentrons.hardware_control import API, ThreadManager
from opentrons.config import (feature_flags as ff, name,
                              robot_configs, IS_ROBOT, ROBOT_FIRMWARE_DIR)
from opentrons.util import logging_config
from opentrons.protocols.types import ApiDeprecationError
from opentrons.protocols.api_support.types import APIVersion

version = sys.version_info[0:2]
if version < (3, 7):
    raise RuntimeError(
        'opentrons requires Python 3.7 or above, this is {0}.{1}'.format(
            version[0], version[1]))

HERE = os.path.abspath(os.path.dirname(__file__))

try:
    with open(os.path.join(HERE, 'package.json')) as pkg:
        package_json = json.load(pkg)
        __version__ = package_json.get('version')
except (FileNotFoundError, OSError):
    __version__ = 'unknown'

from opentrons import config  # noqa: E402

LEGACY_MODULES = [
    'robot', 'reset', 'instruments', 'containers', 'labware', 'modules']

__all__ = ['version', 'HERE', 'config']


def __getattr__(attrname):
    """
    Prevent import of legacy modules from global to officially
    deprecate Python API Version 1.0.
    """
    if attrname in LEGACY_MODULES:
        raise ApiDeprecationError(APIVersion(1, 0))
    raise AttributeError(attrname)


def __dir__():
    return sorted(__all__ + LEGACY_MODULES)


log = logging.getLogger(__name__)

try:
    import systemd.daemon  # type: ignore
    systemdd_notify = partial(systemd.daemon.notify, "READY=1")
except ImportError:
    log.info("Systemd couldn't be imported, not notifying")
    systemdd_notify = partial(lambda: None)


SMOOTHIE_HEX_RE = re.compile('smoothie-(.*).hex')


def _find_smoothie_file() -> Tuple[Path, str]:
    resources: List[Path] = []

    # Search for smoothie files in /usr/lib/firmware first then fall back to
    # value packed in wheel
    if IS_ROBOT:
        resources.extend(ROBOT_FIRMWARE_DIR.iterdir())  # type: ignore

    resources_path = Path(HERE) / 'resources'
    resources.extend(resources_path.iterdir())

    for path in resources:
        matches = SMOOTHIE_HEX_RE.search(path.name)
        if matches:
            branch_plus_ref = matches.group(1)
            return path, branch_plus_ref
    raise OSError(f"Could not find smoothie firmware file in {resources_path}")


async def initialize_robot() -> ThreadManager:
    """Build the hardware controller."""
    if os.environ.get("ENABLE_VIRTUAL_SMOOTHIE"):
        log.info("Initialized robot using virtual Smoothie")
        systemdd_notify()
        return ThreadManager(API.build_hardware_simulator)

    # Check if smoothie emulator is to be used
    port = os.environ.get("OT_SMOOTHIE_EMULATOR_URI")
    if not port:
        smoothie_id = os.environ.get('OT_SMOOTHIE_ID', 'AMA')
        # Let this raise an exception.
        port = get_ports_by_name(device_name=smoothie_id)[0]

    log.info(f"Connecting to smoothie at port {port}")

    packed_smoothie_fw_file, packed_smoothie_fw_ver = _find_smoothie_file()
    systemdd_notify()
    hardware = ThreadManager(API.build_hardware_controller,
                             threadmanager_nonblocking=True,
                             port=port,
                             firmware=(packed_smoothie_fw_file,
                                       packed_smoothie_fw_ver))
    try:
        await hardware.managed_thread_ready_async()
    except RuntimeError:
        log.exception(
            'Could not build hardware controller, forcing virtual')
        return ThreadManager(API.build_hardware_simulator)

    loop = asyncio.get_event_loop()

    async def blink():
        while True:
            await hardware.set_lights(button=True)
            await asyncio.sleep(0.5)
            await hardware.set_lights(button=False)
            await asyncio.sleep(0.5)

    blink_task = loop.create_task(blink())

    if not ff.disable_home_on_boot():
        log.info("Homing Z axes")
        await hardware.home_z()

    blink_task.cancel()
    await hardware.set_lights(button=True)

    return hardware


async def initialize() \
        -> ThreadManager:
    """
    Initialize the Opentrons hardware returning a hardware instance.
    """
    robot_conf = robot_configs.load()
    logging_config.log_init(robot_conf.log_level)

    log.info(f"API server version:  {__version__}")
    log.info(f"Robot Name: {name()}")

    hardware = await initialize_robot()

    return hardware



import asyncio
import logging
from threading import Thread, Event
from typing import Mapping, Union, Optional
from opentrons.drivers.temp_deck import (
    SimulatingDriver, TempDeck as TempDeckDriver)
from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.drivers.temp_deck.driver import temp_locks
from ..execution_manager import ExecutionManager
from . import update, mod_abc, types

log = logging.getLogger(__name__)

TEMP_POLL_INTERVAL_SECS = 1

FIRST_GEN2_REVISION = 20


def _model_from_revision(revision: Optional[str]) -> str:
    """ Defines the revision -> model mapping"""
    if not revision or 'v' not in revision:
        log.error(f'bad revision: {revision}')
        return 'temperatureModuleV1'
    try:
        revision_num = float(revision.split('v')[-1])  # type: ignore
    except (ValueError, TypeError):
        # none or corrupt
        log.exception('no revision')
        return 'temperatureModuleV1'

    if revision_num < FIRST_GEN2_REVISION:
        return 'temperatureModuleV1'
    else:
        return 'temperatureModuleV2'


class MissingDevicePortError(Exception):
    pass


class Poller(Thread):
    def __init__(self, driver: Union[TempDeckDriver, SimulatingDriver]):
        self._driver_ref = driver
        self._stop_event = Event()
        super().__init__(target=self._poll_temperature,
                         name='Temperature poller for tempdeck')

    def _poll_temperature(self):
        while not self._stop_event.wait(TEMP_POLL_INTERVAL_SECS):
            self._driver_ref.update_temperature()

    def stop(self):
        self._stop_event.set()


class TempDeck(mod_abc.AbstractModule):
    """
    Under development. API subject to change without a version bump
    """
    @classmethod
    async def build(cls,
                    port: str,
                    usb_port: USBPort,
                    execution_manager: ExecutionManager,
                    interrupt_callback: types.InterruptCallback = None,
                    simulating: bool = False,
                    loop: asyncio.AbstractEventLoop = None,
                    sim_model: str = None):

        """ Build and connect to a TempDeck"""
        # TempDeck does not currently use interrupts, so the callback is not
        # passed on
        mod = cls(port=port,
                  usb_port=usb_port,
                  simulating=simulating,
                  loop=loop,
                  execution_manager=execution_manager,
                  sim_model=sim_model)
        await mod._connect()
        return mod

    @classmethod
    def name(cls) -> str:
        return 'tempdeck'

    def model(self) -> str:
        return _model_from_revision(self._device_info.get('model'))

    @classmethod
    def bootloader(cls) -> mod_abc.UploadFunction:
        return update.upload_via_avrdude

    @staticmethod
    def _build_driver(
            simulating: bool,
            sim_model: str = None
    ) -> Union['SimulatingDriver', 'TempDeckDriver']:
        if simulating:
            return SimulatingDriver(sim_model=sim_model)
        else:
            return TempDeckDriver()

    def __init__(self,
                 port: str,
                 usb_port: USBPort,
                 execution_manager: ExecutionManager,
                 simulating: bool,
                 loop: asyncio.AbstractEventLoop = None,
                 sim_model: str = None) -> None:
        super().__init__(port=port,
                         usb_port=usb_port,
                         simulating=simulating,
                         loop=loop,
                         execution_manager=execution_manager,
                         sim_model=sim_model)
        self._device_info: Mapping[str, str] = {}
        self._driver: Union['SimulatingDriver', 'TempDeckDriver']
        if temp_locks.get(port):
            self._driver = temp_locks[port][1]
        else:
            self._driver = self._build_driver(
                simulating, sim_model)

        self._poller: Optional[Poller] = None

    async def set_temperature(self, celsius: float):
        """
        Set temperature in degree Celsius
        Range: 4 to 95 degree Celsius (QA tested).
        The internal temp range is -9 to 99 C, which is limited by the 2-digit
        temperature display. Any input outside of this range will be clipped
        to the nearest limit
        """
        await self.wait_for_is_running()
        task = self._loop.create_task(self._driver.set_temperature(celsius))
        await self.make_cancellable(task)
        return await task

    async def start_set_temperature(self, celsius):
        """
        Set temperature in degree Celsius
        Range: 4 to 95 degree Celsius (QA tested).
        The internal temp range is -9 to 99 C, which is limited by the 2-digit
        temperature display. Any input outside of this range will be clipped
        to the nearest limit
        """
        await self.wait_for_is_running()
        return self._driver.start_set_temperature(celsius)

    async def await_temperature(self, awaiting_temperature: float):
        """
        Await temperature in degree Celsius
        Polls temperature module's temperature until
        the specified temperature is reached
        """
        if self.is_simulated:
            return

        await self.wait_for_is_running()

        async def _await_temperature(awaiting_temperature: float):
            status = self.status

            if status == 'heating':
                while self.temperature < awaiting_temperature:
                    await asyncio.sleep(0.2)

            elif status == 'cooling':
                while self.temperature > awaiting_temperature:
                    await asyncio.sleep(0.2)

        t = self._loop.create_task(_await_temperature(awaiting_temperature))
        await self.make_cancellable(t)
        await t

    async def deactivate(self):
        """ Stop heating/cooling and turn off the fan """
        await self.wait_for_is_running()
        self._driver.deactivate()

    @property
    def device_info(self) -> Mapping[str, str]:
        return self._device_info

    @property
    def live_data(self) -> types.LiveData:
        return {
            'status': self.status,
            'data': {
                'currentTemp': self.temperature,
                'targetTemp': self.target
            }
        }

    @property
    def temperature(self) -> float:
        return self._driver.temperature

    @property
    def target(self) -> Optional[float]:
        return self._driver.target

    @property
    def status(self) -> str:
        return self._driver.status

    @property
    def port(self) -> str:
        return self._port

    @property
    def usb_port(self) -> USBPort:
        return self._usb_port

    @property
    def is_simulated(self) -> bool:
        return isinstance(self._driver, SimulatingDriver)

    @property
    def interrupt_callback(self) -> types.InterruptCallback:
        return lambda x: None

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    def set_loop(self, loop: asyncio.AbstractEventLoop):
        self._loop = loop

    async def _connect(self):
        """
        Connect to the 'TempDeck' port
        Planned change- will connect to the correct port in case of multiple
        TempDecks
        """
        if self._poller:
            self._poller.stop()
            self._poller.join()
        if not self._driver.is_connected():
            self._driver.connect(self._port)
        self._device_info = self._driver.get_device_info()
        self._poller = Poller(self._driver)
        self._poller.start()

    def cleanup(self) -> None:
        if hasattr(self, '_poller') and self._poller:
            log.debug("Stopping tempdeck poller.")
            self._poller.stop()

    def __del__(self):
        self.cleanup()

    async def prep_for_update(self) -> str:
        model = self._device_info and self._device_info.get('model')
        if model in ('temp_deck_v1', 'temp_deck_v1.1', 'temp_deck_v2'):
            raise types.UpdateError("This Temperature Module can't be updated."
                                    "Please contact Opentrons Support.")

        if self._poller:
            self._poller.stop()
            self._poller.join()
        del self._poller
        self._poller = None
        self._driver.enter_programming_mode()
        new_port = await update.find_bootloader_port()
        return new_port or self.port

    def has_available_update(self) -> bool:
        """ Override of abc implementation to suppress update notifications
        for v1, v1.1, and v2 temperature modules which cannot be updated """
        if not self._device_info:
            model = None
        else:
            model = self._device_info.get('model')
        if model in {'temp_deck_v1', 'temp_deck_v1.1', 'temp_deck_v2', None}:
            return False
        return super().has_available_update()



from __future__ import annotations

from typing_extensions import Literal, Final, TypedDict
from typing import Optional, List, Sequence, TYPE_CHECKING, Union, Any
from opentrons.hardware_control.modules import ThermocyclerStep

if TYPE_CHECKING:
    from opentrons.protocol_api import InstrumentContext
    from opentrons.protocol_api.labware import Well

from opentrons.types import Location


# type for subscriptions
COMMAND: Final = 'command'

# Robot #

DELAY: Final = 'command.DELAY'
HOME: Final = 'command.HOME'
PAUSE: Final = 'command.PAUSE'
RESUME: Final = 'command.RESUME'
COMMENT: Final = 'command.COMMENT'

# Pipette #

ASPIRATE: Final = 'command.ASPIRATE'
DISPENSE: Final = 'command.DISPENSE'
MIX: Final = 'command.MIX'
CONSOLIDATE: Final = 'command.CONSOLIDATE'
DISTRIBUTE: Final = 'command.DISTRIBUTE'
TRANSFER: Final = 'command.TRANSFER'
PICK_UP_TIP: Final = 'command.PICK_UP_TIP'
DROP_TIP: Final = 'command.DROP_TIP'
BLOW_OUT: Final = 'command.BLOW_OUT'
AIR_GAP: Final = 'command.AIR_GAP'
TOUCH_TIP: Final = 'command.TOUCH_TIP'
RETURN_TIP: Final = 'command.RETURN_TIP'
MOVE_TO: Final = 'command.MOVE_TO'

# Modules #

MAGDECK_CALIBRATE: Final = 'command.MAGDECK_CALIBRATE'
MAGDECK_DISENGAGE: Final = 'command.MAGDECK_DISENGAGE'
MAGDECK_ENGAGE: Final = 'command.MAGDECK_ENGAGE'

TEMPDECK_DEACTIVATE: Final = 'command.TEMPDECK_DEACTIVATE'
TEMPDECK_SET_TEMP: Final = 'command.TEMPDECK_SET_TEMP'
TEMPDECK_AWAIT_TEMP: Final = 'command.TEMPDECK_AWAIT_TEMP'

THERMOCYCLER_OPEN: Final = 'command.THERMOCYCLER_OPEN'
THERMOCYCLER_CLOSE: Final = 'command.THERMOCYCLER_CLOSE'
THERMOCYCLER_SET_BLOCK_TEMP: Final = 'command.THERMOCYCLER_SET_BLOCK_TEMP'
THERMOCYCLER_EXECUTE_PROFILE: Final = 'command.THERMOCYCLER_EXECUTE_PROFILE'
THERMOCYCLER_DEACTIVATE: Final = 'command.THERMOCYCLER_DEACTIVATE'
THERMOCYCLER_WAIT_FOR_HOLD: Final = 'command.THERMOCYCLER_WAIT_FOR_HOLD'
THERMOCYCLER_WAIT_FOR_TEMP: Final = 'command.THERMOCYCLER_WAIT_FOR_TEMP'
THERMOCYCLER_WAIT_FOR_LID_TEMP: Final = 'command.THERMOCYCLER_WAIT_FOR_LID_TEMP'
THERMOCYCLER_SET_LID_TEMP: Final = 'command.THERMOCYCLER_SET_LID_TEMP'
THERMOCYCLER_DEACTIVATE_LID: Final = 'command.THERMOCYCLER_DEACTIVATE_LID'
THERMOCYCLER_DEACTIVATE_BLOCK: Final = 'command.THERMOCYCLER_DEACTIVATE_BLOCK'


class TextOnlyPayload(TypedDict):
    text: str


class SingleLocationPayload(TypedDict):
    location: Union[Location, Well]


class MultiLocationPayload(TypedDict):
    locations: Sequence[Union[Location, Well]]


class OptionalMultiLocationPayload(TypedDict):
    locations: Optional[Sequence[Union[Location, Well]]]


class OptionalSingleLocationPayload(TypedDict):
    location: Union[Location, Well, None]


HasLocationPayload = Union[
    SingleLocationPayload, MultiLocationPayload,
    OptionalSingleLocationPayload, OptionalMultiLocationPayload]


class SingleInstrumentPayload(TypedDict):
    instrument: InstrumentContext


class MultiInstrumentPayload(TypedDict):
    instruments: Sequence[InstrumentContext]


HasInstrumentPayload = Union[SingleInstrumentPayload, MultiInstrumentPayload]


class CommentCommandPayload(TextOnlyPayload):
    pass


class CommentCommand(TypedDict):
    name: Literal['command.COMMENT']
    payload: CommentCommandPayload


class DelayCommandPayload(TextOnlyPayload):
    minutes: float
    seconds: float


class DelayCommand(TypedDict):
    name: Literal['command.DELAY']
    payload: DelayCommandPayload


class PauseCommandPayload(TextOnlyPayload):
    userMessage: Optional[str]


class PauseCommand(TypedDict):
    name: Literal['command.PAUSE']
    payload: PauseCommandPayload


class ResumeCommandPayload(TextOnlyPayload):
    pass


class ResumeCommand(TypedDict):
    name: Literal['command.RESUME']
    payload: ResumeCommandPayload


class MagdeckEngageCommandPayload(TextOnlyPayload):
    pass


class MagdeckEngageCommand(TypedDict):
    name: Literal['command.MAGDECK_ENGAGE']
    payload: MagdeckEngageCommandPayload


class MagdeckDisengageCommandPayload(TextOnlyPayload):
    pass


class MagdeckDisengageCommand(TypedDict):
    name: Literal['command.MAGDECK_DISENGAGE']
    payload: MagdeckDisengageCommandPayload


class MagdeckCalibrateCommandPayload(TextOnlyPayload):
    pass


class MagdeckCalibrateCommand(TypedDict):
    name: Literal['command.MAGDECK_CALIBRATE']
    payload: MagdeckCalibrateCommandPayload


class TempdeckSetTempCommandPayload(TextOnlyPayload):
    celsius: float


class TempdeckSetTempCommand(TypedDict):
    name: Literal['command.TEMPDECK_SET_TEMP']
    payload: TempdeckSetTempCommandPayload


class TempdeckAwaitTempCommandPayload(TextOnlyPayload):
    celsius: float


class TempdeckAwaitTempCommand(TypedDict):
    name: Literal['command.TEMPDECK_AWAIT_TEMP']
    payload: TempdeckAwaitTempCommandPayload


class TempdeckDeactivateCommandPayload(TextOnlyPayload):
    pass


class TempdeckDeactivateCommand(TypedDict):
    name: Literal['command.TEMPDECK_DEACTIVATE']
    payload: TempdeckDeactivateCommandPayload


class ThermocyclerOpenCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerOpenCommand(TypedDict):
    name: Literal['command.THERMOCYCLER_OPEN']
    payload: ThermocyclerOpenCommandPayload


class ThermocyclerSetBlockTempCommandPayload(TextOnlyPayload):
    temperature: float
    hold_time: Optional[float]


class ThermocyclerSetBlockTempCommand(TypedDict):
    name: Literal['command.THERMOCYCLER_SET_BLOCK_TEMP']
    payload: ThermocyclerSetBlockTempCommandPayload


class ThermocyclerExecuteProfileCommandPayload(TextOnlyPayload):
    steps: List[ThermocyclerStep]


class ThermocyclerExecuteProfileCommand(TypedDict):
    name: Literal['command.THERMOCYCLER_EXECUTE_PROFILE']
    payload: ThermocyclerExecuteProfileCommandPayload


class ThermocyclerWaitForHoldCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerWaitForHoldCommand(TypedDict):
    name: Literal['command.THERMOCYCLER_WAIT_FOR_HOLD']
    payload: ThermocyclerWaitForHoldCommandPayload


class ThermocyclerWaitForTempCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerWaitForTempCommand(TypedDict):
    name: Literal['command.THERMOCYCLER_WAIT_FOR_TEMP']
    payload: ThermocyclerWaitForTempCommandPayload


class ThermocyclerSetLidTempCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerSetLidTempCommand(TypedDict):
    name: Literal['command.THERMOCYCLER_SET_LID_TEMP']
    payload: ThermocyclerSetLidTempCommandPayload


class ThermocyclerDeactivateLidCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerDeactivateLidCommand(TypedDict):
    name: Literal['command.THERMOCYCLER_DEACTIVATE_LID']
    payload: ThermocyclerDeactivateLidCommandPayload


class ThermocyclerDeactivateBlockCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerDeactivateBlockCommand(TypedDict):
    name: Literal['command.THERMOCYCLER_DEACTIVATE_BLOCK']
    payload: ThermocyclerDeactivateBlockCommandPayload


class ThermocyclerDeactivateCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerDeactivateCommand(TypedDict):
    name: Literal['command.THERMOCYCLER_DEACTIVATE']
    payload: ThermocyclerDeactivateCommandPayload


class ThermocyclerWaitForLidTempCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerWaitForLidTempCommand(TypedDict):
    name: Literal['command.THERMOCYCLER_WAIT_FOR_LID_TEMP']
    payload: ThermocyclerWaitForLidTempCommandPayload


class ThermocyclerCloseCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerCloseCommand(TypedDict):
    name: Literal['command.THERMOCYCLER_CLOSE']
    payload: ThermocyclerCloseCommandPayload


class HomeCommandPayload(TextOnlyPayload):
    axis: str


class HomeCommand(TypedDict):
    name: Literal['command.HOME']
    payload: HomeCommandPayload


class SimpleLHCommandPayload(
        TextOnlyPayload, SingleLocationPayload, SingleInstrumentPayload):
    volume: float
    rate: float


class SimplePairedLHCommandPayload(
        TextOnlyPayload, MultiLocationPayload, MultiInstrumentPayload):
    volume: float
    rate: float


class AspirateCommand(TypedDict):
    name: Literal['command.ASPIRATE']
    payload: Union[SimpleLHCommandPayload, SimplePairedLHCommandPayload]


class DispenseCommand(TypedDict):
    name: Literal['command.DISPENSE']
    payload: Union[SimpleLHCommandPayload, SimplePairedLHCommandPayload]


class ConsolidateCommandPayload(
        TextOnlyPayload, MultiLocationPayload, SingleInstrumentPayload):
    volume: Union[float, List[float]]
    source: List[Union[Location, Well]]
    dest: Union[Location, Well]


class ConsolidateCommand(TypedDict):
    name: Literal['command.CONSOLIDATE']
    payload: ConsolidateCommandPayload


class DistributeCommandPayload(
        TextOnlyPayload, MultiLocationPayload, SingleInstrumentPayload):
    volume: Union[float, List[float]]
    source: Union[Location, Well]
    dest: List[Union[Location, Well]]


class DistributeCommand(TypedDict):
    name: Literal['command.DISTRIBUTE']
    payload: DistributeCommandPayload


class TransferCommandPayload(
        TextOnlyPayload, MultiLocationPayload, SingleInstrumentPayload):
    volume: Union[float, List[float]]
    source: List[Union[Location, Well]]
    dest: List[Union[Location, Well]]


class TransferCommand(TypedDict):
    name: Literal['command.TRANSFER']
    payload: TransferCommandPayload


class MixCommandPayload(
        TextOnlyPayload, OptionalSingleLocationPayload, SingleInstrumentPayload):
    volume: float
    repetitions: int


class PairedMixCommandPayload(
        TextOnlyPayload, OptionalMultiLocationPayload, MultiInstrumentPayload):
    volume: float
    repetitions: int


class MixCommand(TypedDict):
    name: Literal['command.MIX']
    payload: Union[MixCommandPayload, PairedMixCommandPayload]


class BlowOutCommandPayload(
        TextOnlyPayload, OptionalSingleLocationPayload, SingleInstrumentPayload):
    pass


class PairedBlowOutCommandPayload(
        TextOnlyPayload, OptionalMultiLocationPayload, MultiInstrumentPayload):
    pass


class BlowOutCommand(TypedDict):
    name: Literal['command.BLOW_OUT']
    payload: Union[BlowOutCommandPayload, PairedBlowOutCommandPayload]


class TouchTipCommandPayload(
        TextOnlyPayload, SingleInstrumentPayload):
    pass


class PairedTouchTipCommandPayload(
        TextOnlyPayload, OptionalMultiLocationPayload, MultiInstrumentPayload):
    pass


class TouchTipCommand(TypedDict):
    name: Literal['command.TOUCH_TIP']
    payload: Union[TouchTipCommandPayload, PairedTouchTipCommandPayload]


class AirGapCommandPayload(TextOnlyPayload):
    pass


class AirGapCommand(TypedDict):
    name: Literal['command.AIR_GAP']
    payload: AirGapCommandPayload


class ReturnTipCommandPayload(TextOnlyPayload):
    pass


class ReturnTipCommand(TypedDict):
    name: Literal['command.RETURN_TIP']
    payload: ReturnTipCommandPayload


class PickUpTipCommandPayload(
        TextOnlyPayload, SingleLocationPayload, SingleInstrumentPayload):
    pass


class PairedPickUpTipCommandPayload(
        TextOnlyPayload, MultiLocationPayload, MultiInstrumentPayload):
    pass


class PickUpTipCommand(TypedDict):
    name: Literal['command.PICK_UP_TIP']
    payload: Union[PickUpTipCommandPayload, PairedPickUpTipCommandPayload]
    pass


class DropTipCommandPayload(
        TextOnlyPayload, SingleLocationPayload, SingleInstrumentPayload):
    pass


class PairedDropTipCommandPayload(
        TextOnlyPayload, MultiLocationPayload, MultiInstrumentPayload):
    pass


class DropTipCommand(TypedDict):
    name: Literal['command.DROP_TIP']
    payload: Union[DropTipCommandPayload, PairedDropTipCommandPayload]


class MoveToCommand(TypedDict):
    name: Literal['command.MOVE_TO']
    payload: Union[MoveToCommandPayload, PairedMoveToCommandPayload]


class MoveToCommandPayload(
        TextOnlyPayload, SingleLocationPayload, SingleInstrumentPayload):
    pass


class PairedMoveToCommandPayload(
        TextOnlyPayload, MultiLocationPayload, MultiInstrumentPayload):
    pass


Command = Union[
    DropTipCommand, PickUpTipCommand, ReturnTipCommand, AirGapCommand,
    TouchTipCommand, BlowOutCommand, MixCommand, TransferCommand,
    DistributeCommand, ConsolidateCommand, DispenseCommand, AspirateCommand,
    HomeCommand, ThermocyclerCloseCommand, ThermocyclerWaitForLidTempCommand,
    ThermocyclerDeactivateCommand, ThermocyclerDeactivateBlockCommand,
    ThermocyclerDeactivateLidCommand, ThermocyclerSetLidTempCommand,
    ThermocyclerWaitForTempCommand, ThermocyclerWaitForHoldCommand,
    ThermocyclerExecuteProfileCommand, ThermocyclerSetBlockTempCommand,
    ThermocyclerOpenCommand,
    TempdeckDeactivateCommand, TempdeckAwaitTempCommand,
    TempdeckSetTempCommand, MagdeckCalibrateCommand, MagdeckDisengageCommand,
    MagdeckEngageCommand, ResumeCommand, PauseCommand, DelayCommand,
    CommentCommand, MoveToCommand]


CommandPayload = Union[
    CommentCommandPayload, ResumeCommandPayload,
    MagdeckEngageCommandPayload, MagdeckDisengageCommandPayload,
    MagdeckCalibrateCommandPayload,
    ThermocyclerOpenCommandPayload,
    ThermocyclerWaitForHoldCommandPayload,
    ThermocyclerWaitForTempCommandPayload,
    ThermocyclerSetLidTempCommandPayload,
    ThermocyclerDeactivateLidCommandPayload,
    ThermocyclerDeactivateBlockCommandPayload,
    ThermocyclerDeactivateCommandPayload,
    ThermocyclerWaitForLidTempCommand,
    ThermocyclerCloseCommandPayload,
    AirGapCommandPayload, ReturnTipCommandPayload,
    DropTipCommandPayload, PairedDropTipCommandPayload,
    PickUpTipCommandPayload, PairedPickUpTipCommandPayload,
    TouchTipCommandPayload, PairedTouchTipCommandPayload,
    BlowOutCommandPayload, PairedBlowOutCommandPayload,
    MixCommandPayload, PairedMixCommandPayload,
    TransferCommandPayload, DistributeCommandPayload, ConsolidateCommandPayload,
    SimpleLHCommandPayload, SimplePairedLHCommandPayload,
    HomeCommandPayload,
    ThermocyclerExecuteProfileCommandPayload,
    ThermocyclerSetBlockTempCommandPayload,
    TempdeckAwaitTempCommandPayload,
    TempdeckSetTempCommandPayload,
    PauseCommandPayload, DelayCommandPayload,
    MoveToCommandPayload, PairedMoveToCommandPayload
]


MessageSequenceId = Union[Literal['before'], Literal['after']]


CommandMessageSequence = TypedDict('CommandMessageSequence',

                                   {'$': MessageSequenceId})
CommandMessageMeta = TypedDict('CommandMessageMeta',
                               {'meta': Any}, total=False)


class CommandMessageFields(CommandMessageMeta, CommandMessageSequence):
    pass


class MoveToMessage(CommandMessageFields, MoveToCommand):
    pass


class DropTipMessage(CommandMessageFields, DropTipCommand):
    pass


class PickUpTipMessage(CommandMessageFields, PickUpTipCommand):
    pass


class ReturnTipMessage(CommandMessageFields, ReturnTipCommand):
    pass


class AirGapMessage(CommandMessageFields, AirGapCommand):
    pass


class TouchTipMessage(CommandMessageFields, TouchTipCommand):
    pass


class BlowOutMessage(CommandMessageFields, BlowOutCommand):
    pass


class MixMessage(CommandMessageFields, MixCommand):
    pass


class TransferMessage(CommandMessageFields, TransferCommand):
    pass


class DistributeMessage(CommandMessageFields, DistributeCommand):
    pass


class ConsolidateMessage(CommandMessageFields, ConsolidateCommand):
    pass


class DispenseMessage(CommandMessageFields, DispenseCommand):
    pass


class AspirateMessage(CommandMessageFields, AspirateCommand):
    pass


class HomeMessage(CommandMessageFields, HomeCommand):
    pass


class ThermocyclerCloseMessage(
        CommandMessageFields, ThermocyclerCloseCommand):
    pass


class ThermocyclerWaitForLidTempMessage(
        CommandMessageFields, ThermocyclerWaitForLidTempCommand):
    pass


class ThermocyclerDeactivateMessage(
        CommandMessageFields, ThermocyclerDeactivateCommand):
    pass


class ThermocyclerDeactivateBlockMessage(
        CommandMessageFields, ThermocyclerDeactivateBlockCommand):
    pass


class ThermocyclerDeactivateLidMessage(
        CommandMessageFields, ThermocyclerDeactivateLidCommand):
    pass


class ThermocyclerSetLidTempMessage(
        CommandMessageFields, ThermocyclerSetLidTempCommand):
    pass


class ThermocyclerWaitForTempMessage(
        CommandMessageFields, ThermocyclerWaitForTempCommand):
    pass


class ThermocyclerWaitForHoldMessage(
        CommandMessageFields, ThermocyclerWaitForHoldCommand):
    pass


class ThermocyclerExecuteProfileMessage(
        CommandMessageFields, ThermocyclerExecuteProfileCommand):
    pass


class ThermocyclerSetBlockTempMessage(
        CommandMessageFields, ThermocyclerSetBlockTempCommand):
    pass


class ThermocyclerOpenMessage(
        CommandMessageFields, ThermocyclerOpenCommand):
    pass


class TempdeckDeactivateMessage(
        CommandMessageFields, TempdeckDeactivateCommand):
    pass


class TempdeckAwaitTempMessage(
        CommandMessageFields, TempdeckAwaitTempCommand):
    pass


class TempdeckSetTempMessage(
        CommandMessageFields, TempdeckSetTempCommand):
    pass


class MagdeckCalibrateMessage(
        CommandMessageFields, MagdeckCalibrateCommand):
    pass


class MagdeckDisengageMessage(
        CommandMessageFields, MagdeckDisengageCommand):
    pass


class MagdeckEngageMessage(CommandMessageFields, MagdeckEngageCommand):
    pass


class ResumeMessage(CommandMessageFields, ResumeCommand):
    pass


class PauseMessage(CommandMessageFields, PauseCommand):
    pass


class DelayMessage(CommandMessageFields, DelayCommand):
    pass


class CommentMessage(CommandMessageFields, CommentCommand):
    pass


CommandMessage = Union[
    DropTipMessage, PickUpTipMessage, ReturnTipMessage, AirGapMessage,
    TouchTipMessage, BlowOutMessage, MixMessage, TransferMessage,
    DistributeMessage, ConsolidateMessage, DispenseMessage, AspirateMessage,
    HomeMessage, ThermocyclerCloseMessage, ThermocyclerWaitForLidTempMessage,
    ThermocyclerDeactivateMessage, ThermocyclerDeactivateBlockMessage,
    ThermocyclerDeactivateLidMessage, ThermocyclerSetLidTempMessage,
    ThermocyclerWaitForTempMessage, ThermocyclerWaitForHoldMessage,
    ThermocyclerExecuteProfileMessage, ThermocyclerSetBlockTempMessage,
    ThermocyclerOpenMessage, TempdeckSetTempMessage, TempdeckDeactivateMessage,
    MagdeckEngageMessage, MagdeckDisengageMessage, MagdeckCalibrateMessage,
    CommentMessage, DelayMessage, PauseMessage, ResumeMessage, MoveToMessage]



""" opentrons.protocol_api.module_geometry: classes and functions for modules
as deck objects

This module provides things like :py:class:`ModuleGeometry` and
:py:func:`load_module` to create and manipulate module objects as geometric
objects on the deck (as opposed to calling commands on them, which is handled
by :py:mod:`.module_contexts`)
"""
import functools
import logging
import re
from typing import (Mapping, Optional,
                    Union, TYPE_CHECKING)

import numpy as np  # type: ignore
import jsonschema  # type: ignore
from opentrons.protocols.context.protocol_api.labware import\
    LabwareImplementation

from opentrons_shared_data import module
from opentrons.hardware_control.modules.types import (
    ModuleModel, ModuleType, MagneticModuleModel,
    TemperatureModuleModel, ThermocyclerModuleModel)
from opentrons.types import Location, Point, LocationLabware
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.definitions import (
    MAX_SUPPORTED_VERSION, V2_MODULE_DEF_VERSION)
from opentrons.protocols.geometry.deck_item import DeckItem
from opentrons.protocol_api.labware import Labware

from .types import GenericConfiguration, ThermocyclerConfiguration

if TYPE_CHECKING:
    from opentrons_shared_data.module.dev_types import (
        ModuleDefinitionV1, ModuleDefinitionV2)


def module_model_from_string(model_string: str) -> ModuleModel:
    for model_enum in {
            MagneticModuleModel,
            TemperatureModuleModel,
            ThermocyclerModuleModel}:
        try:
            return model_enum.from_str(model_string)  # type: ignore
        except AttributeError:
            pass
    raise AttributeError(f"No such module model {model_string}")


log = logging.getLogger(__name__)


class NoSuchModuleError(Exception):
    def __init__(self, message: str, requested_module: ModuleModel) -> None:
        self.message = message
        self.requested_module = requested_module
        super().__init__()

    def __str__(self) -> str:
        return self.message


class ModuleGeometry(DeckItem):
    """
    This class represents an active peripheral, such as an Opentrons Magnetic
    Module, Temperature Module or Thermocycler Module. It defines the physical
    geometry of the device (primarily the offset that modifies the position of
    the labware mounted on top of it).
    """

    @property
    def separate_calibration(self) -> bool:
        # If a module is the parent of a labware it affects calibration
        return True

    def __init__(self,
                 display_name: str,
                 model: ModuleModel,
                 module_type: ModuleType,
                 offset: Point,
                 overall_height: float,
                 height_over_labware: float,
                 parent: Location,
                 api_level: APIVersion) -> None:
        """
        Create a Module for tracking the position of a module.

        Note that modules do not currently have a concept of calibration apart
        from calibration of labware on top of the module. The practical result
        of this is that if the module parent :py:class:`.Location` is
        incorrect, then acorrect calibration of one labware on the deck would
        be incorrect on the module, and vice-versa. Currently, the way around
        this would be to correct the :py:class:`.Location` so that the
        calibrated labware is targeted accurately in both positions.

        :param display_name: A human-readable display name of only the module
                             (for instance, "Thermocycler Module" - not
                             including parents or location)
        :param model: The model of module this represents
        :param offset: The offset from the slot origin at which labware loaded
                       on this module should be placed
        :param overall_height: The height of the module without labware
        :param height_over_labware: The height of this module over the top of
                                    the labware
        :param parent: A location representing location of the front left of
                       the outside of the module (usually the front-left corner
                       of a slot on the deck).
        :type parent: :py:class:`.Location`
        :param APIVersion api_level: the API version to set for the loaded
                                     :py:class:`ModuleGeometry` instance. The
                                     :py:class:`ModuleGeometry` will
                                     conform to this level.
        """
        self._api_version = api_level
        self._parent = parent
        self._module_type = module_type
        self._display_name = "{} on {}".format(
            display_name, str(parent.labware))
        self._model = model
        self._offset = offset
        self._height = overall_height + self._parent.point.z
        self._over_labware = height_over_labware
        self._labware: Optional[Labware] = None
        self._location = Location(
            point=self._offset + self._parent.point,
            labware=self)

    @property
    def api_version(self) -> APIVersion:
        return self._api_version

    def add_labware(self, labware: Labware) -> Labware:
        assert not self._labware,\
            '{} is already on this module'.format(self._labware)
        self._labware = labware
        return self._labware

    def reset_labware(self):
        self._labware = None

    @property
    def model(self) -> ModuleModel:
        return self._model

    @property
    def load_name(self) -> str:
        return self.model.value

    @property
    def module_type(self) -> ModuleType:
        """ The Moduletype """
        return self._module_type

    @property
    def parent(self) -> LocationLabware:
        return self._parent.labware.object

    @property
    def labware(self) -> Optional[Labware]:
        return self._labware

    @property
    def location(self) -> Location:
        """
        :return: a :py:class:`.Location` representing the top of the module
        """
        return self._location

    @property
    def labware_offset(self) -> Point:
        """
        :return: a :py:class:`.Point` representing the transformation
        between the critical point of the module and the critical
        point of its contained labware
        """
        return self._offset

    @property
    def highest_z(self) -> float:
        if self.labware:
            return self.labware.highest_z + self._over_labware
        else:
            return self._height

    def __repr__(self) -> str:
        return self._display_name


class ThermocyclerGeometry(ModuleGeometry):
    def __init__(self,
                 display_name: str,
                 model: ModuleModel,
                 module_type: ModuleType,
                 offset: Point,
                 overall_height: float,
                 height_over_labware: float,
                 lid_height: float,
                 parent: Location,
                 api_level: APIVersion,
                 configuration: GenericConfiguration =
                 ThermocyclerConfiguration.FULL) -> None:
        """
        Create a Module for tracking the position of a module.

        Note that modules do not currently have a concept of calibration apart
        from calibration of labware on top of the module. The practical result
        of this is that if the module parent :py:class:`.Location` is
        incorrect, then acorrect calibration of one labware on the deck would
        be incorrect on the module, and vice-versa. Currently, the way around
        this would be to correct the :py:class:`.Location` so that the
        calibrated labware is targeted accurately in both positions.

        :param display_name: A human-readable display name of only the module
                             (for instance, "Thermocycler Module" - not
                             including parents or location)
        :param model: The model of module this represents
        :param offset: The offset from the slot origin at which labware loaded
                       on this module should be placed
        :param overall_height: The height of the module without labware
        :param height_over_labware: The height of this module over the top of
                                    the labware
        :param parent: A location representing location of the front left of
                       the outside of the module (usually the front-left corner
                       of a slot on the deck).
        :type parent: :py:class:`.Location`
        :param APIVersion api_level: the API version to set for the loaded
                                     :py:class:`ModuleGeometry` instance. The
                                     :py:class:`ModuleGeometry` will
                                     conform to this level.
        :param configuration: Used to specify the slot configuration of
                              the Thermocycler. It should by of type
                              :py:class:`.ThermocyclerConfiguration` and can
                              either be FULL or SEMI.
        """
        super().__init__(
            display_name, model, module_type, offset, overall_height,
            height_over_labware, parent, api_level)
        self._lid_height = lid_height
        self._lid_status = 'open'   # Needs to reflect true status
        self._configuration = configuration
        if self.is_semi_configuration:
            self._offset = self._offset + Point(-23.28, 0, 0)

    @property
    def highest_z(self) -> float:
        # TODO: BC 2019-08-27 this highest_z value represents the distance
        # from the top of the open TC chassis to the base. Once we have a
        # more robust collision detection system in place, the collision
        # model for the TC should change based on it's lid_status
        # (open or closed). A prerequisite for that check will be
        # path-specific highest z calculations, as opposed to the current
        # global check on instrument.move_to. For example: a move from slot 1
        # to slot 3 should only check the highest z of all deck items between
        # the source and destination in the x,y plane.
        return super().highest_z

    @property
    def lid_status(self) -> str:
        return self._lid_status

    @lid_status.setter
    def lid_status(self, status) -> None:
        self._lid_status = status

    @property
    def is_semi_configuration(self) -> bool:
        return bool(self._configuration == ThermocyclerConfiguration.SEMI)

    @property
    def covered_slots(self) -> set:
        if self.is_semi_configuration:
            return {7, 10}
        else:
            return {7, 8, 10, 11}

    # NOTE: this func is unused until "semi" configuration
    def labware_accessor(self, labware: Labware) -> Labware:
        # Block first three columns from being accessed
        definition = labware._implementation.get_definition()
        definition['ordering'] = definition['ordering'][2::]
        return Labware(
            implementation=LabwareImplementation(definition, super().location),
            api_level=self._api_version
        )

    def add_labware(self, labware: Labware) -> Labware:
        assert not self._labware,\
            '{} is already on this module'.format(self._labware)
        assert self.lid_status != 'closed', \
            'Cannot place labware in closed module'
        if self.is_semi_configuration:
            self._labware = self.labware_accessor(labware)
        else:
            self._labware = labware
        return self._labware


def _load_from_v1(definition: 'ModuleDefinitionV1',
                  parent: Location,
                  api_level: APIVersion) -> ModuleGeometry:
    """ Load a module geometry from a v1 definition.

    The definition should be schema checked before being passed to this
    function; all definitions passed here are assumed to be valid.
    """
    mod_name = definition['loadName']
    model_lookup: Mapping[str, ModuleModel] = {
        'thermocycler': ThermocyclerModuleModel.THERMOCYCLER_V1,
        'magdeck': MagneticModuleModel.MAGNETIC_V1,
        'tempdeck': TemperatureModuleModel.TEMPERATURE_V1}
    type_lookup = {
        'thermocycler': ModuleType.THERMOCYCLER,
        'tempdeck': ModuleType.TEMPERATURE,
        'magdeck': ModuleType.MAGNETIC
    }
    model = model_lookup[mod_name]
    offset = Point(definition["labwareOffset"]["x"],
                   definition["labwareOffset"]["y"],
                   definition["labwareOffset"]["z"])
    overall_height = definition["dimensions"]["bareOverallHeight"]\

    height_over_labware = definition["dimensions"]["overLabwareHeight"]

    if model in ThermocyclerModuleModel:
        lid_height = definition['dimensions']['lidHeight']
        mod: ModuleGeometry = \
            ThermocyclerGeometry(definition["displayName"],
                                 model,
                                 type_lookup[mod_name],
                                 offset,
                                 overall_height,
                                 height_over_labware,
                                 lid_height,
                                 parent,
                                 api_level)
    else:
        mod = ModuleGeometry(definition['displayName'],
                             model,
                             type_lookup[mod_name],
                             offset,
                             overall_height,
                             height_over_labware,
                             parent, api_level)
    return mod


def _load_from_v2(definition: 'ModuleDefinitionV2',
                  parent: Location,
                  api_level: APIVersion,
                  configuration: GenericConfiguration) -> ModuleGeometry:
    """ Load a module geometry from a v2 definition.

    The definition should be schema checked before being passed to this
     function; all definitions passed here are assumed to be valid.
    """
    pre_transform = np.array((definition['labwareOffset']['x'],
                              definition['labwareOffset']['y'],
                              1))
    if not parent.labware.is_slot:
        par = ''
        log.warning(
            f'module location parent labware was {parent.labware} which is'
            'not a slot name; slot transforms will not be loaded')
    else:
        par = str(parent.labware.object)

    # this needs to change to look up the current deck type if/when we add
    # that notion
    xforms_ser = definition['slotTransforms']\
        .get('ot2_standard', {})\
        .get(par, {'labwareOffset': [[1, 0, 0], [0, 1, 0], [0, 0, 1]]})
    xform_ser = xforms_ser['labwareOffset']

    # apply the slot transform if any
    xform = np.array(xform_ser)
    xformed = np.dot(xform, pre_transform)
    if definition['moduleType'] in {
            ModuleType.MAGNETIC.value,
            ModuleType.TEMPERATURE.value}:
        return ModuleGeometry(
            parent=parent,
            api_level=api_level,
            offset=Point(xformed[0],
                         xformed[1],
                         definition['labwareOffset']['z']),
            overall_height=definition['dimensions']['bareOverallHeight'],
            height_over_labware=definition['dimensions']['overLabwareHeight'],
            model=module_model_from_string(definition['model']),
            module_type=ModuleType.from_str(definition['moduleType']),
            display_name=definition['displayName'])
    elif definition['moduleType'] == ModuleType.THERMOCYCLER.value:
        return ThermocyclerGeometry(
            parent=parent,
            api_level=api_level,
            offset=Point(xformed[0],
                         xformed[1],
                         definition['labwareOffset']['z']),
            overall_height=definition['dimensions']['bareOverallHeight'],
            height_over_labware=definition['dimensions']['overLabwareHeight'],
            model=module_model_from_string(definition['model']),
            module_type=ModuleType.from_str(definition['moduleType']),
            display_name=definition['displayName'],
            lid_height=definition['dimensions']['lidHeight'],
            configuration=configuration)
    else:
        raise RuntimeError(f'Unknown module type {definition["moduleType"]}')


def load_module_from_definition(
        definition: Union['ModuleDefinitionV1', 'ModuleDefinitionV2'],
        parent: Location,
        api_level: APIVersion = None,
        configuration: GenericConfiguration =
        ThermocyclerConfiguration.FULL) -> ModuleGeometry:
    """
    Return a :py:class:`ModuleGeometry` object from a specified definition
    matching the v1 module definition schema

    :param definition: A dict representing the full module definition adhering
                       to the v1 module schema
    :param parent: A :py:class:`.Location` representing the location where
                   the front and left most point of the outside of the module
                   is (often the front-left corner of a slot on the deck).
    :param APIVersion api_level: the API version to set for the loaded
                                 :py:class:`ModuleGeometry` instance. The
                                 :py:class:`ModuleGeometry` will
                                 conform to this level. If not specified,
                                 defaults to :py:attr:`.MAX_SUPPORTED_VERSION`.
    """
    api_level = api_level or MAX_SUPPORTED_VERSION
    # def not yet discriminated, mypy complains sadly
    schema = definition.get("$otSharedSchema")  # type: ignore
    if not schema:
        # v1 definitions don't have schema versions
        # but unfortunately we can't tell mypy that because it can only
        # discriminate unions based on literal tags or similar, not the
        # presence or absence of keys
        v1def: 'ModuleDefinitionV1' = definition  # type: ignore
        return _load_from_v1(v1def, parent, api_level)
    if schema == 'module/schemas/2':
        schema_doc = module.load_schema('2')
        try:
            jsonschema.validate(definition, schema_doc)
        except jsonschema.ValidationError:
            log.exception("Failed to validate module def schema")
            raise RuntimeError('The specified module definition is not valid.')
        # mypy can't tell these apart, but we've schema validated it - this is
        # the right type
        v2def: 'ModuleDefinitionV2' = definition  # type: ignore
        return _load_from_v2(v2def, parent, api_level, configuration)
    elif isinstance(schema, str):
        maybe_schema = re.match('^module/schemas/([0-9]+)$', schema)
        if maybe_schema:
            raise RuntimeError(
                f"Module definitions of schema version {maybe_schema.group(1)}"
                " are not supported in this robot software release.")
    log.error(f"Bad module definition (schema specifier {schema})")
    raise RuntimeError(
        'The specified module definition is not valid.')


def _load_v1_module_def(module_model: ModuleModel) -> 'ModuleDefinitionV1':
    v1names = {MagneticModuleModel.MAGNETIC_V1: 'magdeck',
               TemperatureModuleModel.TEMPERATURE_V1: 'tempdeck',
               ThermocyclerModuleModel.THERMOCYCLER_V1: 'thermocycler'}
    try:
        name = v1names[module_model]
    except KeyError:
        raise NoSuchModuleError(
            f'Could not find module {module_model.value}',
            module_model)
    return module.load_definition('1', name)


def _load_v2_module_def(module_model: ModuleModel) -> 'ModuleDefinitionV2':
    try:
        return module.load_definition('2', module_model.value)
    except module.ModuleNotFoundError:
        raise NoSuchModuleError(
            f'Could not find the module {module_model.value}.',
            module_model)


@functools.lru_cache(maxsize=128)
def _load_module_definition(
        api_level: APIVersion,
        module_model: ModuleModel) -> Union['ModuleDefinitionV2',
                                            'ModuleDefinitionV1']:
    """
    Load the appropriate module definition for this api version
    """

    if api_level < V2_MODULE_DEF_VERSION:
        try:
            return _load_v1_module_def(module_model)
        except NoSuchModuleError:
            try:
                dname = _load_v2_module_def(module_model)['displayName']
            except NoSuchModuleError:
                dname = module_model.value
            raise NoSuchModuleError(
                f'API version {api_level} does not support the module '
                f'{dname}. Please use at least version '
                f'{V2_MODULE_DEF_VERSION} to use this module.', module_model)
    else:
        return _load_v2_module_def(module_model)


def load_module(
        model: ModuleModel,
        parent: Location,
        api_level: APIVersion = None,
        configuration: str = None) -> ModuleGeometry:
    """
    Return a :py:class:`ModuleGeometry` object from a definition looked up
    by name.

    :param model: The module model to use. This should be one of the strings
                  returned by :py:func:`ModuleGeometry.resolve_module_model`
    :param parent: A :py:class:`.Location` representing the location where
                   the front and left most point of the outside of the module
                   is (often the front-left corner of a slot on the deck).
    :param APIVersion api_level: the API version to set for the loaded
                                 :py:class:`ModuleGeometry` instance. The
                                 :py:class:`ModuleGeometry` will
                                 conform to this level. If not specified,
                                 defaults to :py:attr:`.MAX_SUPPORTED_VERSION`.
    :param configuration: Used to specify the slot configuration of
                        the Thermocycler. Only Valid in Python API
                        Version 2.4 and later. If you wish to use
                        the non-full plate configuration, you must
                        pass in the key word value `semi`
    """
    api_level = api_level or MAX_SUPPORTED_VERSION
    defn = _load_module_definition(api_level, model)
    if configuration:
        # Here we are converting the string passed in to a
        # configuration type. For now we only have
        # Thermocycler configurations, but there could be others
        # in the future
        return load_module_from_definition(
            defn, parent, api_level,
            ThermocyclerConfiguration.configuration_type(configuration))
    else:
        return load_module_from_definition(defn, parent, api_level)


def resolve_module_model(module_name: str) -> ModuleModel:
    """ Turn any of the supported load names into module model names """

    model_map: Mapping[str, ModuleModel] = {
        'magneticModuleV1': MagneticModuleModel.MAGNETIC_V1,
        'magneticModuleV2': MagneticModuleModel.MAGNETIC_V2,
        'temperatureModuleV1': TemperatureModuleModel.TEMPERATURE_V1,
        'temperatureModuleV2': TemperatureModuleModel.TEMPERATURE_V2,
        'thermocyclerModuleV1': ThermocyclerModuleModel.THERMOCYCLER_V1,
    }

    alias_map: Mapping[str, ModuleModel] = {
        'magdeck': MagneticModuleModel.MAGNETIC_V1,
        'magnetic module': MagneticModuleModel.MAGNETIC_V1,
        'magnetic module gen2': MagneticModuleModel.MAGNETIC_V2,
        'tempdeck': TemperatureModuleModel.TEMPERATURE_V1,
        'temperature module': TemperatureModuleModel.TEMPERATURE_V1,
        'temperature module gen2': TemperatureModuleModel.TEMPERATURE_V2,
        'thermocycler': ThermocyclerModuleModel.THERMOCYCLER_V1,
        'thermocycler module': ThermocyclerModuleModel.THERMOCYCLER_V1,
    }

    lower_name = module_name.lower()
    resolved_name = model_map.get(module_name, None) \
        or alias_map.get(lower_name, None)
    if not resolved_name:
        raise ValueError(
            f'{module_name} is not a valid module load name.\n'
            'Valid names (ignoring case): ''"' + '", "'
            .join(alias_map.keys()) + '"\n' +
            'You can also refer to modules by their ' +
            'exact model: ''"' + '", "'
            .join(model_map.keys()) + '"'
            )
    return resolved_name


def resolve_module_type(module_model: ModuleModel) -> ModuleType:
    return ModuleType.from_str(_load_v2_module_def(module_model)['moduleType'])


def models_compatible(model_a: ModuleModel, model_b: ModuleModel) -> bool:
    """ Check if two module models may be considered the same """
    if model_a == model_b:
        return True
    return model_b.value in _load_v2_module_def(model_a)['compatibleWith']



from .definition import (
    get_labware_definition, get_all_labware_definitions, save_calibration,
    save_definition, get_labware_hash, get_labware_hash_with_parent,
    delete_all_custom_labware, verify_definition)
from .load import load, load_from_definition


__all__ = [
    "load",
    "load_from_definition",
    "get_labware_definition",
    "get_all_labware_definitions",
    "load_from_definition",
    "save_definition",
    "verify_definition",
    "delete_all_custom_labware",
    "save_calibration",
    "get_labware_hash",
    "get_labware_hash_with_parent",
]



from . import args_handler

BAD_BARCODE_MESSAGE = 'Unexpected Serial -> {}'
WRITE_FAIL_MESSAGE = 'Data not saved, HOLD BUTTON'

# must iterate through v1.4 and v1.3 first, because v1 barcodes did not
# have characters to specify the version number
VERSIONS = ['v2.2', 'v2.1', 'v2', 'v1.6', 'v1.5', 'v1.4', 'v1.3', 'v1']

MODELS = {
    'v1': {
        'P10S': 'p10_single_v1',
        'P10M': 'p10_multi_v1',
        'P50S': 'p50_single_v1',
        'P50M': 'p50_multi_v1',
        'P300S': 'p300_single_v1',
        'P300M': 'p300_multi_v1',
        'P1000S': 'p1000_single_v1'
    },
    'v1.3': {
        'P10SV13': 'p10_single_v1.3',
        'P10MV13': 'p10_multi_v1.3',
        'P50SV13': 'p50_single_v1.3',
        'P50MV13': 'p50_multi_v1.3',
        'P3HSV13': 'p300_single_v1.3',
        'P3HMV13': 'p300_multi_v1.3',
        'P1KSV13': 'p1000_single_v1.3'
    },
    'v1.4': {
        'P10SV14': 'p10_single_v1.4',
        'P10MV14': 'p10_multi_v1.4',
        'P50SV14': 'p50_single_v1.4',
        'P50MV14': 'p50_multi_v1.4',
        'P3HSV14': 'p300_single_v1.4',
        'P3HMV14': 'p300_multi_v1.4',
        'P1KSV14': 'p1000_single_v1.4'
    },
    'v1.5': {
        'P10SV15': 'p10_single_v1.5',
        'P10MV15': 'p10_multi_v1.5',
        'P50SV15': 'p50_single_v1.5',
        'P50MV15': 'p50_multi_v1.5',
        'P3HSV15': 'p300_single_v1.5',
        'P3HMV15': 'p300_multi_v1.5',
        'P1KSV15': 'p1000_single_v1.5'
    },
    'v1.6': {
        'P10MV16': 'p10_multi_v1.6',
    },
    'v2': {
        'P3HSV20': 'p300_single_v2.0',
        'P3HMV20': 'p300_multi_v2.0',
        'P1KSV20': 'p1000_single_v2.0',
        'P20SV20': 'p20_single_v2.0',
        'P20MV20': 'p20_multi_v2.0',
    },
    'v2.1': {
        'P3HSV21': 'p300_single_v2.1',
        'P1KSV21': 'p1000_single_v2.1',
        'P20SV21': 'p20_single_v2.1',
        'P3HMV21': 'p300_multi_v2.1',
        'P20MV21': 'p20_multi_v2.1',
    },
    'v2.2': {
        'P20MV22': 'p20_multi_v2.2',
        'P20SV22': 'p20_single_v2.2',
        'P1KSV22': 'p1000_single_v2.2',
    }
}


def write_identifiers(mount, new_id, new_model, driver):
    '''
    Send a bytearray to the specified mount, so that Smoothieware can
    save the bytes to the pipette's memory
    '''
    driver.write_pipette_id(mount, new_id)
    read_id = driver.read_pipette_id(mount)
    _assert_the_same(new_id, read_id)
    driver.write_pipette_model(mount, new_model)
    read_model = driver.read_pipette_model(mount)
    _assert_the_same(new_model, read_model)


def check_previous_data(mount, driver):
    old_id = driver.read_pipette_id(mount)
    old_model = driver.read_pipette_model(mount)
    if old_id and old_model:
        print(
            'Overwriting old data: id={0}, model={1}'.format(
                old_id, old_model))
    else:
        print('No old data on this pipette')


def _assert_the_same(a, b):
    if a != b:
        raise Exception(WRITE_FAIL_MESSAGE)


def _user_submitted_barcode(max_length):
    '''
    User can enter a serial number as a string of HEX values
    Length of byte array must equal `num`
    '''
    barcode = input('BUTTON + SCAN: ').strip()
    if len(barcode) > max_length:
        raise Exception(BAD_BARCODE_MESSAGE.format(barcode))
    # remove all characters before the letter P
    # for example, remove ASCII selector code "\x1b(B" on chinese keyboards
    barcode = barcode[barcode.index('P'):]
    barcode = barcode.split('\r')[0].split('\n')[0]  # remove any newlines
    return barcode


def _parse_model_from_barcode(barcode):
    for version in VERSIONS:
        for barcode_substring in MODELS[version].keys():
            if barcode.startswith(barcode_substring):
                return MODELS[version][barcode_substring]
    raise Exception(BAD_BARCODE_MESSAGE.format(barcode))


def _do_wpm(driver):
    try:
        barcode = _user_submitted_barcode(32)
        model = _parse_model_from_barcode(barcode)
        check_previous_data('right', driver)
        write_identifiers('right', barcode, model, driver)
        print('PASS: Saved -> {0} (model {1})'.format(barcode, model))
    except KeyboardInterrupt:
        exit()
    except Exception as e:
        print('FAIL: {}'.format(e))


def main():
    parser = args_handler.root_argparser(
        "Write model and serial to a pipette's eeprom")
    args = parser.parse_args()
    _, driver = args_handler.build_driver(args.port)
    while True:
        _do_wpm(driver)


if __name__ == "__main__":
    main()



from typing import Dict, Iterable

from opentrons.protocol_engine import WellLocation, WellOrigin
from opentrons.protocol_engine.commands import (
    CommandRequestType, AspirateRequest, DispenseRequest,
    PickUpTipRequest, DropTipRequest
)
from opentrons.protocols.models import json_protocol as models


class CommandTranslatorError(Exception):
    pass


ReturnType = Iterable[CommandRequestType]


class CommandTranslator:
    """Class that translates commands from PD/JSON to ProtocolEngine."""

    def __init__(self) -> None:
        """Construct a command translator"""
        pass

    def translate(self, command: models.AllCommands) -> ReturnType:
        """
        Translate a PD/JSON protocol to Protocol Engine commands.

        Args:
            command: The PD/JSON command to translate

        Returns:
            A collection of Protocol Engine command requests.

        """
        try:
            h = CommandTranslator._COMMAND_TO_NAME[command.command]
            return getattr(self, h)(command)
        except KeyError:
            raise CommandTranslatorError(
                f"'{command.command}' is not recognized.")
        except AttributeError:
            raise CommandTranslatorError(
                f"Cannot find handler for '{command.command}'.")

    def _aspirate(
            self,
            command: models.LiquidCommand) -> ReturnType:
        """
        Translate an aspirate JSON command to a protocol engine aspirate request.

        Args:
            command: JSON protocol aspirate command

        Returns: AspirateRequest

        """
        # TODO (al, 2021-04-26): incoming pipette and labware ids are
        #  assigned by PD. Are they the same as Protocol Engine's?
        return [
            AspirateRequest(
                pipetteId=command.params.pipette,
                labwareId=command.params.labware,
                wellName=command.params.well,
                volume=command.params.volume,
                wellLocation=WellLocation(
                    origin=WellOrigin.BOTTOM,
                    offset=(0, 0, command.params.offsetFromBottomMm)
                )
            )
        ]

    def _dispense(
            self,
            command: models.LiquidCommand) -> ReturnType:
        """
        Translate a dispense JSON command to a protocol engine dispense request.

        Args:
            command: JSON protocol dispense command

        Returns: DispenseRequest
        """
        return [
            DispenseRequest(
                pipetteId=command.params.pipette,
                labwareId=command.params.labware,
                wellName=command.params.well,
                volume=command.params.volume,
                wellLocation=WellLocation(
                    origin=WellOrigin.BOTTOM,
                    offset=(0, 0, command.params.offsetFromBottomMm)
                )
            )
        ]

    def _air_gap(
            self,
            command: models.LiquidCommand) -> ReturnType:
        raise NotImplementedError()

    def _blowout(
            self,
            command: models.BlowoutCommand) -> ReturnType:
        raise NotImplementedError()

    def _touch_tip(
            self,
            command: models.TouchTipCommand) -> ReturnType:
        raise NotImplementedError()

    def _pick_up(
            self,
            command: models.PickUpDropTipCommand) -> ReturnType:
        """
        Translate a pick_up_tip JSON command to a protocol engine pick_up_tip request.

        Args:
            command: JSON protocol PickUpTip command

        Returns: PickUpTipRequest
        """
        return [
            PickUpTipRequest(
                pipetteId=command.params.pipette,
                labwareId=command.params.labware,
                wellName=command.params.well
            )
        ]

    def _drop_tip(
            self,
            command: models.PickUpDropTipCommand) -> ReturnType:
        """
        Translate a drop tip JSON command to a protocol engine drop tip request.

        Args:
            command: JSON protocol drop tip command

        Returns: DropTipRequest

        """
        return [
            DropTipRequest(
                pipetteId=command.params.pipette,
                labwareId=command.params.labware,
                wellName=command.params.well
            )
        ]

    def _move_to_slot(
            self,
            command: models.MoveToSlotCommand) -> ReturnType:
        raise NotImplementedError()

    def _delay(
            self,
            command: models.DelayCommand) -> ReturnType:
        raise NotImplementedError()

    def _magnetic_module_engage(
            self, command: models.MagneticModuleEngageCommand) -> ReturnType:
        raise NotImplementedError()

    def _magnetic_module_disengage(
            self,
            command: models.MagneticModuleDisengageCommand) -> ReturnType:
        raise NotImplementedError()

    def _temperature_module_set_target(
            self,
            command: models.TemperatureModuleSetTargetCommand) -> ReturnType:
        raise NotImplementedError()

    def _temperature_module_await_temperature(
            self,
            command: models.TemperatureModuleAwaitTemperatureCommand) -> ReturnType:
        raise NotImplementedError()

    def _temperature_module_deactivate(
            self,
            command: models.TemperatureModuleDeactivateCommand) -> ReturnType:
        raise NotImplementedError()

    def _thermocycler_set_target_block_temperature(
            self,
            command: models.ThermocyclerSetTargetBlockTemperatureCommand) -> ReturnType:
        raise NotImplementedError()

    def _thermocycler_set_target_lid_temperature(
            self,
            command: models.ThermocyclerSetTargetLidTemperatureCommand) -> ReturnType:
        raise NotImplementedError()

    def _thermocycler_await_block_temperature(
            self,
            command: models.ThermocyclerAwaitBlockTemperatureCommand) -> ReturnType:
        raise NotImplementedError()

    def _thermocycler_await_lid_temperature(
            self,
            command: models.ThermocyclerAwaitLidTemperatureCommand) -> ReturnType:
        raise NotImplementedError()

    def _thermocycler_deactivate_block(
            self,
            command: models.ThermocyclerDeactivateBlockCommand) -> ReturnType:
        raise NotImplementedError()

    def _thermocycler_deactivate_lid(
            self,
            command: models.ThermocyclerDeactivateLidCommand) -> ReturnType:
        raise NotImplementedError()

    def _thermocycler_open_lid(
            self,
            command: models.ThermocyclerOpenLidCommand) -> ReturnType:
        raise NotImplementedError()

    def _thermocycler_close_lid(
            self,
            command: models.ThermocyclerCloseLidCommand) -> ReturnType:
        raise NotImplementedError()

    def _thermocycler_run_profile(
            self,
            command: models.ThermocyclerRunProfile) -> ReturnType:
        raise NotImplementedError()

    def _thermocycler_await_profile_complete(
            self,
            command: models.ThermocyclerAwaitProfileCompleteCommand) -> ReturnType:
        raise NotImplementedError()

    def _move_to_well(
            self,
            command: models.MoveToWellCommand) -> ReturnType:
        raise NotImplementedError()

    _COMMAND_TO_NAME: Dict[str, str] = {
        models.CommandMoveToWell:
            _move_to_well.__name__,
        models.CommandThermocyclerAwaitProfile:
            _thermocycler_await_profile_complete.__name__,
        models.CommandThermocyclerRunProfile:
            _thermocycler_run_profile.__name__,
        models.CommandThermocyclerCloseLid:
            _thermocycler_close_lid.__name__,
        models.CommandThermocyclerOpenLid:
            _thermocycler_open_lid.__name__,
        models.CommandThermocyclerDeactivateLid:
            _thermocycler_deactivate_lid.__name__,
        models.CommandThermocyclerDeactivateBlock:
            _thermocycler_deactivate_block.__name__,
        models.CommandThermocyclerSetTargetLid:
            _thermocycler_set_target_lid_temperature.__name__,
        models.CommandThermocyclerAwaitBlockTemperature:
            _thermocycler_await_block_temperature.__name__,
        models.CommandThermocyclerAwaitLidTemperature:
            _thermocycler_await_lid_temperature.__name__,
        models.CommandThermocyclerSetTargetBlock:
            _thermocycler_set_target_block_temperature.__name__,
        models.CommandTemperatureModuleDeactivate:
            _temperature_module_deactivate.__name__,
        models.CommandTemperatureModuleAwait:
            _temperature_module_await_temperature.__name__,
        models.CommandTemperatureModuleSetTarget:
            _temperature_module_set_target.__name__,
        models.CommandMagneticModuleDisengage:
            _magnetic_module_disengage.__name__,
        models.CommandMagneticModuleEngage:
            _magnetic_module_engage.__name__,
        models.CommandDelay:
            _delay.__name__,
        models.CommandMoveToSlot:
            _move_to_slot.__name__,
        models.CommandDropTip:
            _drop_tip.__name__,
        models.CommandPickUpTip:
            _pick_up.__name__,
        models.CommandTouchTip:
            _touch_tip.__name__,
        models.CommandBlowout:
            _blowout.__name__,
        models.CommandAirGap:
            _air_gap.__name__,
        models.CommandDispense:
            _dispense.__name__,
        models.CommandAspirate:
            _aspirate.__name__,
    }



"""Drop tip command request, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel

from .command import CommandImplementation, CommandHandlers
from .pipetting_common import BasePipettingRequest


class DropTipRequest(BasePipettingRequest):
    """A request to drop a tip in a specific well."""

    def get_implementation(self) -> DropTipImplementation:
        """Get the drop tip request's command implementation."""
        return DropTipImplementation(self)


class DropTipResult(BaseModel):
    """Result data from the execution of a DropTipRequest."""

    pass


class DropTipImplementation(
    CommandImplementation[DropTipRequest, DropTipResult]
):
    """Drop tip command implementation."""

    async def execute(self, handlers: CommandHandlers) -> DropTipResult:
        """Move to and drop a tip using the requested pipette."""
        await handlers.pipetting.drop_tip(
            pipette_id=self._request.pipetteId,
            labware_id=self._request.labwareId,
            well_name=self._request.wellName,
        )

        return DropTipResult()



import enum
from itertools import groupby
from dataclasses import dataclass
from typing import List, Optional, Tuple
from opentrons.hardware_control.types import BoardRevision
from opentrons.algorithms.types import GenericNode


class PinDir(enum.Enum):
    rev_input = enum.auto()
    input = enum.auto()
    output = enum.auto()


REV_OG_USB_PORTS = {'3': 1, '5': 2}
REV_A_USB_HUB = 3


@dataclass(frozen=True)
class USBPort(GenericNode):
    port_number: Optional[int] = None
    device_path: str = ''
    hub: Optional[int] = None

    @classmethod
    def build(cls, port_path: str, board_revision: BoardRevision) -> 'USBPort':
        """
        Build a USBPort dataclass.

        An example port path:
        `1-1.3/1-1.3:1.0/tty/ttyACM1/dev`

        :param port_path: Full path of a usb device
        :returns: Tuple of the port number, hub and name
        """
        full_name, device_path = port_path.split(':')
        port_nodes = cls.get_unique_nodes(full_name)
        *port_info, name = cls.find_hub(port_nodes)
        hub, port = cls.map_to_revision(board_revision, port_info)
        return cls(
            name=name,
            port_number=port,
            sub_names=[],
            device_path=device_path,
            hub=hub)

    @staticmethod
    def find_hub(
            port_nodes: List[str]
            ) -> Tuple[Optional[int], int, str]:
        """
        Find Hub.

        Here we need to determine if a port is a hub
        or not. A hub path might look like:
        `1-1.4/1-1.4/1-1.4.1/1-1.4.1`. When this function
        is used, the nodes will look like: ['1-1.4', '1-1.4.1'].
        This function will then be used to check if it is a
        port hub based on the values given. In the case of the
        example above, it will determine this device is
        connected to a hub and return both the port and
        the hub number. The hub would always be the first number,
        in this case `4` and the port number of the hub would be `1`.

        :param port_nodes: A list of unique port id(s)
        :returns: Tuple of the port number, hub and name
        """
        if len(port_nodes) > 1:
            port_info = port_nodes[1].split('.')
            hub: Optional[int] = int(port_info[1])
            port = int(port_info[2])
            name = port_nodes[1]
        else:
            port = int(port_nodes[0].split('.')[1])
            hub = None
            name = port_nodes[0]
        return hub, port, name

    @staticmethod
    def get_unique_nodes(full_name: str) -> List[str]:
        """
        Get Unique Nodes.

        A path might look like: `1-1.3/1-1.3`. In this
        instance we know that the device is on Bus 1 and
        port 3 of the pi. We only need one unique id
        here, so we will filter it out.

        :param full_name: Full path of the physical
        USB Path.
        :returns: List of separated USB port paths
        """
        port_nodes = []
        for node in full_name.split('/'):
            if node not in port_nodes:
                port_nodes.append(node)
        return port_nodes

    @staticmethod
    def map_to_revision(
            board_revision: BoardRevision,
            port_info: List) -> Tuple[Optional[int], int]:
        hub, port = port_info
        if board_revision == BoardRevision.OG:
            if hub:
                return REV_OG_USB_PORTS.get(str(hub), hub), port
            else:
                return hub, REV_OG_USB_PORTS.get(str(port), port)
        else:
            if hub and hub == REV_A_USB_HUB:
                return None, port
            else:
                return hub, port

    def __hash__(self) -> int:
        """
        Hash function.

        To have a unique set of nodes, they must
        all have a unique hash. Lists are not
        hashable which is why we need to unpack
        the list here.
        """
        return hash((self.name, *self.sub_names))


class GPIOPin:

    @classmethod
    def build(cls, name: str, in_out: PinDir, pin: int):
        # use this method if the pin number is the same
        # across all board revisions
        return cls(name, in_out, rev_og=pin,
                   rev_a=pin, rev_b=pin, rev_c=pin)

    @classmethod
    def build_with_rev(cls, name: str, in_out: PinDir,
                       **kwargs):
        return cls(name, in_out, **kwargs)

    def __init__(self, name: str, in_out: PinDir,
                 rev_og: Optional[int] = None,
                 rev_a: Optional[int] = None,
                 rev_b: Optional[int] = None,
                 rev_c: Optional[int] = None):
        self.name = name
        self.in_out = in_out
        self.rev_og = rev_og
        self.rev_a = rev_a
        self.rev_b = rev_b
        self.rev_c = rev_c

    def by_board_rev(self, board_rev: BoardRevision):
        ref = {
            BoardRevision.OG: self.rev_og,
            BoardRevision.A: self.rev_a,
            BoardRevision.B: self.rev_b,
            BoardRevision.C: self.rev_c,
            BoardRevision.UNKNOWN: self.rev_og}
        return ref[board_rev]


class GPIOGroup:
    def __init__(self, pins: List[GPIOPin]):
        self.pins = pins

    def __getattr__(self, item):
        return next(filter(lambda x: x.name is item, self.pins), None)

    def by_type(self, pin_dir: PinDir):
        return GPIOGroup(list(
            filter(lambda x: x.in_out is pin_dir, self.pins)))

    def by_names(self, names: List[str]):
        return GPIOGroup(list(
            filter(lambda x: x.name in names, self.pins)))

    def group_by_pins(self, board_rev: BoardRevision) -> List:
        c = groupby(self.pins, key=lambda x: x.by_board_rev(board_rev))
        l: list = []
        for k, v in c:
            l.append(list(v))
        return l


gpio_group = GPIOGroup([
        # revision pins (input)
        GPIOPin.build('rev_0', PinDir.rev_input, 17),
        GPIOPin.build('rev_1', PinDir.rev_input, 27),
        # output pins
        GPIOPin.build('frame_leds', PinDir.output, 6),
        GPIOPin.build('blue_button', PinDir.output, 13),
        GPIOPin.build('halt', PinDir.output, 18),
        GPIOPin.build('green_button', PinDir.output, 19),
        GPIOPin.build('audio_enable', PinDir.output, 21),
        GPIOPin.build('isp', PinDir.output, 23),
        GPIOPin.build('reset', PinDir.output, 24),
        GPIOPin.build('red_button', PinDir.output, 26),
        # input pins
        GPIOPin.build('button_input', PinDir.input, 5),
        GPIOPin.build_with_rev('door_sw_filt', PinDir.input,
                               rev_og=20, rev_a=12),
        GPIOPin.build_with_rev('window_sw_filt', PinDir.input,
                               rev_og=20, rev_a=16),
        GPIOPin.build('window_door_sw', PinDir.input, 20)])
